"""
This is the zignal library

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2013 Ronny Andersson
@license: MIT
"""

__version__ = "0.3.0"

from .audio import *
from . import filters
from . import measure
from . import music
from . import sndcard

__all__ = [
           'filters',
           'measure',
           'music',
           'sndcard',
           ]
__all__.extend(audio.__all__)       #@UndefinedVariable



"""
Created on 25 Jan 2014

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2014 Ronny Andersson
@license: MIT
"""

# standard library
import logging

# external libraries
import numpy as np
import scipy.signal
import matplotlib.pyplot as plt

# local libraries
from zignal import hz2rad, rad2hz

#===================================================================================================
# Linear filter implementations
#
# Julius O. Smith III - Introduction to Digital Filters with audio applications
# https://ccrma.stanford.edu/~jos/fp/
#
#===================================================================================================
class Filter(object):
    def __init__(self, B=None, A=None, fs=96000):
        """Standard linear recursive filter

                b0 + b1*z^(-1) + b2*z^(-2) + ... + bn*z^(-n)
        H(z) = ------------------------
                a0 + a1*z^(-1) + a2*z^(-2) + ... + an*z^(-n)

        B is feed forward     (numerator)    zeros
        A is feed back        (denominator)  poles
        """
        self._logger = logging.getLogger(__name__)
        assert fs > 0
        self.fs = fs
        self._B = None
        self._A = None

        self.set_coefficients(B=B, A=A)

    def __str__(self):
        s  = '=======================================\n'
        s += 'classname        : %s\n'          %self.__class__.__name__
        s += 'sample rate      : %.1f [Hz]\n'   %self.fs
        s += 'feedforward  (B) : %s\n'          %str(self._B)
        s += 'feedback     (A) : %s\n'          %str(self._A)
        s += 'number of zeros  : %i\n'          %(len(self._B)-1)
        s += 'number of poles  : %i\n'          %(len(self._A)-1)
        s += 'minimum phase?   : %s\n'          %("Yes" if self.is_minimum_phase() else "No")
        s += '-----------------:---------------------\n'
        return s

    def __repr__(self):
        return "Filter(B=%s, A=%s, fs=%s)" %(list(self._B), list(self._A), self.fs)

    def filter_samples(self, samples):
        return scipy.signal.lfilter(self._B, self._A, samples, axis=0)

    def set_coefficients(self, B=None, A=None):
        """Set the filter coefficients.

        B is coefficients in the numerator. This determines the location of the zeros.
        B is also called feedforward coefficients. An FIR filter only has B coefficients
        and will always be stable.

        A is coefficients in the denominator. This determines the location of the poles.
        A is also called feedback coefficients. An IIR filter must have A coefficients
        and might become unstable.
        """

        self._logger.debug("set coefficients (class Filter)")

        # numerator (zeros, feedforward, FIR)
        self._B = np.array((1,)) if B is None else np.array(B)

        # denominator (poles, feedback, IIR)
        self._A = np.array((1,)) if A is None else np.array(A)

        assert len(self._B) != 0
        assert len(self._A) != 0

    def get_coefficients(self):
        return self._B, self._A

    def get_feed_forward(self):
        return self._B

    def get_feed_back(self):
        return self._A

    def normalise(self):
        """Normalise the coefficients by dividing by A[0], effectively
        setting A[0]=1.0
        """
        assert len(self._A) >= 1
        assert len(self._B) >= 1

        a0 = self._A[0]
        self._logger.debug("normalising using a0: %.4f" %a0)

        self._B = self._B/a0
        self._A = self._A/a0

    def is_stable(self):
        """A filter is stable if all its poles are strictly inside the unit circle. A
        pole on the unit circle may be called marginally stable.

        TODO: calculate the reflection coefficients instead, since higher order filters
        are sensitive to coefficient rounding errors in the root finding procedure.
        """
        isStable = True

        unused_zeros, poles, unused_gain = scipy.signal.tf2zpk(self._B, self._A)

        for pole in poles:
            if not np.abs(pole) < 1.0:
                isStable = False

        return isStable

    def is_minimum_phase(self):
        """A filter is defined as _minimum phase_ if all its poles and zeros are inside
        the unit circle, excluding the unit circle itself.
        """
        isMinPhase = True

        zeros, poles, unused_gain = scipy.signal.tf2zpk(self._B, self._A)

        for pole in poles:
            if not np.abs(pole) < 1.0:
                isMinPhase = False

        for zero in zeros:
            if not np.abs(zero) < 1.0:
                isMinPhase = False

        return isMinPhase

    def complex_freq_resp(self, frequencies=None):
        """Calculate the complex frequency response"""
        if frequencies is None:
            w, h = scipy.signal.freqz(self._B, self._A, worN=None)
        elif isinstance(frequencies, int):
            w, h = scipy.signal.freqz(self._B, self._A, worN=frequencies)
        else:
            w, h = scipy.signal.freqz(self._B, self._A, worN=hz2rad(frequencies, self.fs))

        return w, h

    def magnitude_resp(self, frequencies=None):
        """Calculate the real magnitude (frequency) response"""
        w, h    = self.complex_freq_resp(frequencies)
        mag     = 20*np.log10(np.absolute(h))
        freqs   = rad2hz(w, self.fs)

        return freqs, mag

    def phase_resp(self, frequencies=None, unwrap=False):
        """Calculate the real phase response"""
        w, h    = self.complex_freq_resp(frequencies)
        phase   = np.angle(h, deg=False)
        phase   = np.unwrap(phase) if unwrap else phase
        phase   = np.rad2deg(phase)
        freqs   = rad2hz(w, self.fs)

        return freqs, phase

    def plot_mag_phase(self, filename=None, plotpoints=10000, unwrap=False):
        """Produce a plot with magnitude and phase response in the same
        figure. The y-axis on the left side belongs to the magnitude
        response. The y-axis on the right side belongs to the phase
        response
        """
        unused_freq, mag = self.magnitude_resp(plotpoints)
        freq, pha = self.phase_resp(plotpoints, unwrap=unwrap)

        fig = plt.figure(1)
        ax_mag = fig.add_subplot(111)
        ax_pha = ax_mag.twinx()
        ax_mag.semilogx(freq, mag, label='magnitude', color='red',  linestyle='-')
        ax_pha.semilogx(freq, pha, label='phase',     color='blue', linestyle='--')
        ax_mag.grid(True)

        ax_mag.set_xlim(10, self.fs/2)
        #ax_mag.set_ylim(bottom=-80)    # FIXME: ad proper padding
        #ax_mag.margins(0.1)

        ax_mag.set_title('Frequency response')
        ax_mag.set_xlabel('Frequency [Hz]')
        ax_mag.set_ylabel('Magnitude [dB]')
        ax_pha.set_ylabel('Phase [deg]')

        handles1, labels1 = ax_mag.get_legend_handles_labels()
        handles2, labels2 = ax_pha.get_legend_handles_labels()

        plt.legend(handles1 + handles2, labels1 + labels2, loc='best')

        if filename is None:
            plt.show()
        else:
            try:
                plt.savefig(filename)
            finally:
                plt.close(1)

    def plot_pole_zero(self, filename=None):
        """Produce a plot with the location of all poles and zeros."""
        zeros, poles, gain = scipy.signal.tf2zpk(self._B, self._A)
        self._logger.debug("zeros: %s" %zeros)
        self._logger.debug("poles: %s" %poles)
        self._logger.debug("gain : %s" %gain)

        fig = plt.figure(1)
        ax = fig.add_subplot(111, aspect='equal')
        circ = plt.Circle((0,0), radius=1, fill=False, color='black',
                          linestyle='dashed', linewidth=1.0)
        ax.add_patch(circ)
        ax.axhline(0, linestyle='dashed', color='black', linewidth=1.0)
        ax.axvline(0, linestyle='dashed', color='black', linewidth=1.0)
        ax.grid(True)

        ax.plot(poles.real, poles.imag, marker='x', ms=7.0, mew=1.5, mfc='blue', mec='blue',
                ls='None', label='poles (%i)' %len(poles))
        ax.plot(zeros.real, zeros.imag, marker='o', ms=7.0, mew=1.5, mfc='None', mec='red',
                ls='None', label='zeros (%i)' %len(zeros))

        ax.margins(0.1)

        # TODO: count multiples at (0,0)

        plt.legend(loc='best', numpoints=1)
        plt.title('Pole-zero locations')
        plt.xlabel('Real')
        plt.ylabel('Imaginary')

        if filename is None:
            plt.show()
        else:
            try:
                plt.savefig(filename)
            finally:
                plt.close(1)

    def plot_impulse_resp(self, filename=None, points=1000):
        """Plot the impulse response"""
        t, m = scipy.signal.dimpulse((self._B, self._A, 1/self.fs), n=points)
        m = m[0]

        #assert isinstance(m, np.ndarray)

        zoom = np.max(np.abs(m[1:]))

        plt.figure(1)
        plt.plot(t, m, '-x', label='impulse response')
        ax = plt.axis()
        plt.axis([ax[0], ax[1], -zoom*1.25, zoom*1.25])
        plt.grid(True)
        plt.title('Impulse response')
        plt.xlabel('Time [s]')
        plt.ylabel('Magnitude')

        if filename is None:
            plt.show()
        else:
            try:
                plt.savefig(filename)
            finally:
                plt.close(1)

class FIR(Filter):
    """Finite Impulse Response filter

    Also known as
     * nonrecursive filter
     * tapped delay line filter
     * moving average filter
     * transversal filter

    Per definition always stable (no feedback coefficients)
    """
    def __init__(self, B=None, fs=96000):
        Filter.__init__(self, B=B, A=(1,), fs=fs)

    def __str__(self):
        s  = Filter.__str__(self)
        # += '-----------------:---------------------\n'
        s += 'noise amplf.     : %s\n' %self.noise_amplification()
        return s

    def noise_amplification(self):
        '''The noise amplification is the sum of the squares of the coefficients'''
        return np.sum(np.power(self._B, 2))

class IIR(Filter):
    """Infinite Impulse Response

    Also known as
     * recursive filter
     * Ladder filter
     * Lattice filter
     * autoregressive moving average filter (ARMA)
     * autoregressive integrated moving average filter (ARIMA)

    """
    def __init__(self, B=None, A=None, fs=96000):
        Filter.__init__(self, B=B, A=A, fs=fs)

    def __str__(self):
        s  = Filter.__str__(self)
        # += '-----------------:---------------------\n'
        s += 'stable?          : %s\n' %("Yes" if self.is_stable() else "No")
        return s

#===================================================================================================
# Functions
#===================================================================================================

def normalised_frequency(f0=1000, fs=96000):
    '''Calculate a normalised frequency between [0.0, 1.0] where 1.0
    corresponds to pi [rad/sample]
    '''
    return f0/(fs/2)

__all__ = ['Filter', 'FIR', 'IIR', 'normalised_frequency',]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')

    print('++ End of script ++')



"""
Created on 22 Mar 2014

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2014 Ronny Andersson
@license: MIT
"""

# standard library
from collections import deque
import logging
import random

# external libraries
import numpy as np
import scipy.signal

# local libraries
from zignal import Audio
from zignal.filters.biquads import RBJ
from zignal.filters.linearfilter import Filter

# expose the measure.mlstaps.TAPS dictionary in the measure.mls namespace
from .mlstaps import *

__all__ = [
           'MLS',
           'MLS_simple',
           'get_random_taps',
           'TAPS',
           ]

def get_random_taps(N):
    '''Select a random set of taps for given order N'''
    return random.choice(TAPS.get(N))

class _MLS_base(object):
    '''Base class for Maximum Length Sequence (MLS) generation. This class
    implements the "tricky" bits when it comes to MLS sequences. An efficient
    generator is needed since MLS is based on shift registers where we are
    dealing with bits.

    It should not be needed to create an instance of this class directly. Use
    the derived child class instead. When debugging or trying to understand
    the inner workings of MLS then this class can be instantiated with a low
    value for N. For example, N=3 will yield an MLS sequence of length 7. This
    is short enough for a pen and paper exercise when trying to apply circular
    cross correlation.

    The basis of MLS theory is not depending on a sample rate. This is because
    we are dealing with sequences of ones and zeros. This can be translated to
    samples with values -1.0 and +1.0 but unaware of any *sample rate*.

    We can later filter the sequence and apply emphasis. At that stage we need
    to know about sample rate but not in this class.

    This class implements:
    * Generators
    * Circular Cross Correlation
    * Extraction of the impulse response
    '''
    def __init__(self, N=None, taps=None):
        assert N    is not None, "Please specify MLS order"
        assert taps is not None, "Please specify feedback taps"
        assert isinstance(taps, (tuple, list))
        assert len(taps) is not 0, "taps are empty!"

        self._logger    = logging.getLogger(__name__)
        self.N          = N
        self.L          = (2**N)-1
        self.taps       = taps

        sizeof_int = np.int64().dtype.itemsize
        self._RAM_usage = (self.L*sizeof_int)/(1024**2) # in Mb

    def __repr__(self):
        return '_MLS_base(N=%i, taps=%s)' %(self.N, tuple(self.taps))

    def __str__(self):
        s  = '=======================================\n'
        s += 'classname        : %s\n'      %self.__class__.__name__
        s += 'N                : %i\n'      %self.N
        s += 'L=(2^N)-1        : %i\n'      %self.L
        s += 'taps             : %s\n'      %str(self.taps)
        s += "RAM              : %.1f [Mb] (one full sequence)\n" %self._RAM_usage
        s += '-----------------:---------------------\n'
        return s

    def generator_bit(self):
        """Generate one bit at a time. This generator never finishes, it will
        generate bits forever. However, the sequence will repeat itself after
        (2^N)-1 steps.

        The returned value is 1 or 0
        """

        # deque is memory efficient and fast. Here we use it as a shift
        # register with a fixed length.
        shiftregister = deque([0]*(self.N-1)+[1], maxlen=self.N)
        #self._logger.debug("*init* %s" %str(shiftregister))

        while True:
            bitvalue = 0
            for tap in self.taps:
                bitvalue = bitvalue ^ shiftregister[self.N-tap]
            shiftregister.append(bitvalue)

            # The genration of one bit every iteration is costly enough. For debugging
            # purposes, uncomment below to get the internal state of the shiftregister
            # for every iteration, given the conditional N < 6.
            #if self.N < 6:
            #    self._logger.debug("state: %s" %str(shiftregister))
            yield bitvalue

    def generator_chunk(self, chunk=1024):
        """Generate a chunk of values, forever. Wraps around at interval (2^N)-1

        Returns a numpy array of length=chunk with datatype integers.
        """
        bitgenerator = self.generator_bit()

        while True:
            seq = np.zeros((chunk, 1), dtype=np.int64)
            for i in range(chunk):
                seq[i] = next(bitgenerator)
            yield seq

    def generator_samples(self, chunk=1024):
        """Generate a chunk of audio samples in the range [-1.0, 1.0]

        Returns a numpy array. The datatype is 64 bit floats.
        """
        chunkgen = self.generator_chunk(chunk)
        while True:
            binarychunk = next(chunkgen)

            # create a view of the data. This points to the same memory structure.
            # Compare this to the union structure in c++. Data size must match that
            # of the generator_chunk() method (sizeof(np.int64)==sizeof(np.float64))
            samples = binarychunk.view(np.float64)

            # The values that the binarychunk can be is 0 and 1. We want this to represent
            # a sample so we need to convert this to -1.0 and +1.0 instead. We do this by
            # multiplying by two and subtracting 1. The values are now converted so that
            # 0 is -1 and +1 is unchanged. Range is now [-1, 1]. Finally we flip the sign
            # of the sequence.

            samples[:] = -(binarychunk*2 - 1)

            yield samples

    def get_full_sequence(self, repeats=1):
        """Get the full MLS sequence as audio samples, repeated n times. When
        extracting the impulse response we need to throw away the first MLS
        sequence. Hence we need to repeat the signal at least once, i.e
        repeat=2. More repetitions will mean longer signal but better signal
        to noise ratio since we will average away any noise.
        """

        chunkgen        = self.generator_samples(chunk=self.L)
        full_sequence   = next(chunkgen)
        reps_sequence   = np.tile(full_sequence.T, repeats).T
        self._logger.debug("May share memory: %s" %np.may_share_memory(full_sequence,
                                                                       reps_sequence))
        return reps_sequence

    def xcorr_circular(self, other):
        """Circular cross correlation. The difference between circular cross
        correlation and regular correlation is that with circular correlation
        we assume that the signal is repeating.

        The input data "other" has to be the same size as a full sequence.

        Returns the (normalised) impulse response of length L.
        """

        #    Example:
        #
        #    _MLS_base(N=3, taps=(3, 2))
        #
        #    reference  : [[ 1. -1. -1. -1.  1.  1. -1.]]
        #
        #    <do cross-correlation>  <-- in this example this is the auto correlation
        #
        #    xcorr      : [[-1.  2.  1. -2. -3.  0.  7.  0. -3. -2.  1.  2. -1.]]
        #    x1 (view)  :                             [[ 0. -3. -2.  1.  2. -1.]]
        #    x2 (view)  : [[-1.  2.  1. -2. -3.  0.]]
        #
        #    <assume sequence is circular>
        #
        #    x1 (view)  : [[ 0. -3. -2.  1.  2. -1.]]
        #    x2 (view)  : [[-1.  2.  1. -2. -3.  0.]]
        #    x2=x1+x2   :   -1. -1. -1. -1. -1. -1.
        #
        #    because we use a "view" of the array
        #
        #    xcorr      : [[-1.  2.  1. -2. -3.  0.  7. -1. -1. -1. -1. -1. -1.]]
        #
        #    norm       :                         [[ 7. -1. -1. -1. -1. -1. -1.]]
        #
        #    <normalise by L>
        #
        #    norm/L     : [[ 1.     -0.1429 -0.1429 -0.1429 -0.1429 -0.1429 -0.1429]]

        ref = self.get_full_sequence(repeats=1)

        # Correlation and convolution are related. Correlation in the time domain is *very*
        # slow for long sequences. Convolution in the frequecy domain is much faster. We can
        # use the convolution method if we "undo" the flip of the input signal. This is
        # equivalent to the correlation method. There might be some small (insignificant)
        # rounding errors but for a large array this should not be noticable.
        #
        ### Flip comments to verify that correlation and convolution (with input flip) is
        ### the same. Be careful, long sequences are slow to calculate using the
        ### correlation method.
        #xcorr   = scipy.signal.correlate(ref, other)
        xcorr = scipy.signal.fftconvolve(np.flipud(ref), other)

        self._logger.debug("ref: %s" %np.array_str(ref.T,     max_line_width=200, precision=4, suppress_small=True))
        self._logger.debug("xc : %s" %np.array_str(xcorr.T,   max_line_width=200, precision=4, suppress_small=True))

        del ref

        # slicing creates views which are cheap (points to the same array)
        x1 = xcorr[self.L:]         # right half (end)
        x2 = xcorr[:self.L-1]       # left halt (start)
        self._logger.debug("x1 : %s" %np.array_str(x1.T,      max_line_width=200, precision=4, suppress_small=True))
        self._logger.debug("x2 : %s" %np.array_str(x2.T,      max_line_width=200, precision=4, suppress_small=True))

        x1[:]=x1+x2                 # assume circular sequence
        self._logger.debug("xc : %s" %np.array_str(xcorr.T,   max_line_width=200, precision=4, suppress_small=True))

        norm = xcorr[self.L-1:]     # extract "impulse" + tail (right half)
        self._logger.debug("nrm: %s" %np.array_str(norm.T,    max_line_width=200, precision=4, suppress_small=True))

        norm[:] = norm/self.L       # normalise so that max <= 1.0
        self._logger.debug("nrm: %s" %np.array_str(norm.T,    max_line_width=200, precision=4, suppress_small=True))

        return norm

    def trim_and_avgerage(self, other):
        """It is assumed that at least two full MLSes are used and sent through the
        system we want to measure. By using m repeated sequences of MLSes we can throw
        away the first full sequence. This first part might be delayed because of
        latency and will contain the startup response of the system which we aren't
        interested in. We then have n=m-1 sequences. Remember that MLS is a repeated
        sequence and it doesn't matter that we have for example latency of for example
        one third of a sequence. This will show up as phase shift which can be
        compensated for.

        We now calculate the average of the n sequences. This gives us even more noise
        resistance.

        Returns a numpy array of length L.
        """
        # data is [[A_1],
        #          [A_2],
        #          [A_3],
        #          [B_1],
        #          [B_2],
        #          [B_3],
        #          [C_1],
        #          [C_2],
        #          [C_3]]

        # throw away first full sequence
        trimmed = other[self.L:]
        self._logger.debug("May share memory: %s" %np.may_share_memory(other, trimmed))
        #print(repr(trimmed))
        # data is [[B_1],
        #          [B_2],
        #          [B_3],
        #          [C_1],
        #          [C_2],
        #          [C_3]]

        repeats = len(trimmed)//self.L
        self._logger.debug("repeats (first discared): %i" %repeats)

        # reshape so we can average
        reshaped = trimmed.reshape((repeats, self.L))
        # data is [[B_1, B_2, B_3],
        #          [C_1, C_2, C_3]]

        self._logger.debug("May share memory: %s" %np.may_share_memory(other, reshaped))
        #print(repr(reshaped))

        # the averaging finally creates a new array that has it's own data
        average = np.average(reshaped, axis=0)

        average = np.expand_dims(average, axis=1) # up the ndim to 2
        # data is [[X_1],
        #          [X_2],
        #          [X_3]]

        return average

    def get_impulse(self, x):
        """Extract the impulse response by averaging the sequences and then
        applying circular cross correlation. The length of input x must be
        a multiple of L.

        Returns a numpy array with the extracted impulse response.
        """
        assert isinstance(x, np.ndarray)
        assert len(x) > self.L, "The first sequence will be thrown away"

        avg     = self.trim_and_avgerage(x)
        impulse = self.xcorr_circular(avg)

        return impulse

class MLS(_MLS_base, Audio):
    """This class is a mixture of the MLS base class and the Audio class. The MLS data
    is the audio samples in the Audio class. Because we also know about sample rate
    here we can apply emphasis and de-emphasis on the samples. When plotting the FFT
    of the impulse response remember to use a rectangular window. This is perfectly
    valid since the MLS is cyclic by nature and will wrap around.

    This class is limited to one channel of audio. Creating a multichannel MLS signal
    is outside the scope of this class. It can easily be done by creating multiple
    instances and then extracting the audio samples and constructing a new Audio
    instance with all channels appended. To extract the impulse response we then need
    to keep track of which channel is paired with which MLS since they should
    preferably be using different taps to minimise cross talk.
    """
    def __init__(self, N=None, taps=None, fs=96000, repeats=2, B=(1, 0, 0), A=(1, 0, 0)):
        """N is the order of the MLS, taps are preferably selected from the mlstaps
        dictionary, like

            >>> taps=TAPS[N][0]

        B and A are emphasis filter coefficients. The filter used as emphasis must
        be a minimum phase filter. This means that all the poles and the zeroes are
        withing the unit circle. We can then invert the filter to apply de-emphasis.

        The filters.biquads.RBJ class can be used to generate a suitable emphasis
        filter.
        """
        assert repeats > 1, "at least two sequences are needed, (repeats=2)"

        _MLS_base.__init__(self, N=N, taps=taps)
        Audio.__init__(self, fs=fs, initialdata=self.get_full_sequence(repeats=repeats))

        self.repeats            = repeats
        self._length_impresp    = self.L/self.fs
        self._filter_emphasis   = Filter(B=B, A=A, fs=self.fs)
        self._filter_deemphasis = Filter(B=A, A=B, fs=self.fs) # inverse filter

        assert self._filter_emphasis.is_minimum_phase(), \
            "The emphasis filter must be minimum phase, i.e. possible to invert"

    def __repr__(self):
        B, A = self._filter_emphasis.get_coefficients()

        s = 'MLS(N=%i, taps=%s, fs=%r, repeats=%i, B=%s, A=%s)' \
            %(self.N, tuple(self.taps), self.fs, self.repeats, tuple(B), tuple(A))

        return s

    def __str__(self):
        B, A = self._filter_emphasis.get_coefficients()

        mls_string = _MLS_base.__str__(self)
        mls_string = "\n".join(mls_string.splitlines()[2:-1])

        s  = Audio.__str__(self)
        s += '%s\n'                             %mls_string
        s += 'repeats          : %i\n'          %self.repeats
        s += 'len(impulse)     : %.3f [s]\n'    %self._length_impresp
        s += 'emphasis filt. B : %s\n'          %str(B)
        s += 'emphasis filt. A : %s\n'          %str(A)
        return s

    def apply_emphasis(self):
        """Apply emphasis by filtering the whole audio signal. If high freqencies are
        boosted noise will be suppressed. If low frequencies are boosted we will get
        better signal to noise ratio in the lower area of the frequency response.
        """
        self._logger.debug("Applying emphasis filter, in place")
        self.samples = self._filter_emphasis.filter_samples(self.samples)

    def apply_deemphasis(self, x):
        """Undo the emphasis filter by filtering the signal with the inverse
        of the emphasis filter.
        """
        self._logger.debug("Applying de-emphasis filter")
        deemphasis_x = self._filter_deemphasis.filter_samples(x)
        return deemphasis_x

    def get_impulse(self, x):
        """Extract the impulse response. Returns an Audio instance.
        """
        imp = _MLS_base.get_impulse(self, x)
        y   = Audio(fs=self.fs, initialdata=imp)
        return y

class MLS_simple(object):
    def __init__(self, N=16, fs=96000, repeats=3):
        """Simplified usage for quick access to an MLS that performs the required
        message calls in the right order. Can also be used as an example on how
        to use the MLS class.

        Example:
            >>> mls = MLS_simple(N=18, fs=96000, repeats=4)
            >>> y = some_system_to_identify(mls.samples)
            >>> imp = mls.get_impulse(y)
            >>> imp.plot()
            >>> mls.plot_fft()

        The impulse response is stored as a member variable in this class after it
        has been extracted.
        """
        emphasis_filter = RBJ(filtertype="highshelf", gaindb=-10, f0=100, Q=0.707, fs=fs)
        B, A = emphasis_filter.get_coefficients()
        self._mls = MLS(N=N, taps=TAPS[N][0], fs=fs, repeats=repeats, B=B, A=A)
        self._mls.apply_emphasis()

        # map the name to the internal representation of the samples
        self.samples = self._mls.samples

    def __repr__(self):
        s = 'MLS_simple(N=%i fs=%r, repeats=%i)' %(self._mls.N, self._mls.fs, self._mls.repeats)
        return s

    def __str__(self):
        mls_string = str(self._mls)
        mls_string = "\n".join(mls_string.splitlines()[2:])

        s  = '=======================================\n'
        s += 'classname        : %s\n'      %self.__class__.__name__
        s += '%s' %str(mls_string)
        return s

    def get_impulse(self, x):
        """Extract the impulse response"""
        tmp = self._mls.apply_deemphasis(x)
        self._impulseresponse = self._mls.get_impulse(tmp)
        return self._impulseresponse

    def plot_fft(self, plotname=None):
        """Plot the magnitude response. Phase is not included in this plot."""
        assert hasattr(self, "_impulseresponse"), "call get_impulse(...) before trying to plot"

        # A window function on the impulse response does not work here, since
        # that would scale the impulse wrong.
        self._impulseresponse.plot_fft(plotname=plotname, window='rectangular', normalise=False)

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')
    fs  = 48000*2
    N   = 13

    #mls = _MLS_base(N=N, taps=TAPS[N][0])
    #mls = MLS(N=N, taps=TAPS[N][0], fs=fs)

    f = RBJ(filtertype="highshelf", gaindb=-10, f0=100, Q=0.707, fs=fs)
    B, A = f.get_coefficients()
    #mls = MLS(N=N, taps=TAPS[N][0], fs=fs, repeats=5, B=B, A=A)

    mls = MLS_simple(N=N, fs=fs, repeats=4)

    print (repr(mls))
    print(mls)

    y = mls.get_impulse(mls.samples)
    #mls.plot_fft()                                      # used with MLS_simple
    y.plot_fft(window='rectangular', normalise=False)   # used with MLS

    print('++ End of script ++')



'''
Created on Dec 31, 2013

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2013 Ronny Andersson
@license: MIT
'''

# standard library
import logging
import os
import types

# external libraries
import numpy as np
import scipy.signal
import scipy.io.wavfile
import matplotlib.pyplot as plt

#===================================================================================================
# Classes
#===================================================================================================

class Audio(object):
    def __init__(self, channels=0, fs=96000, nofsamples=0, duration=None,
                 initialdata=None, dtype=np.float64):
        """Base class for audio processing. Samples are stored as a numpy array.

        We can create an instance by specifying a channel count and one of either
        a duration or a sample count parameter. The other way of creating an
        instance is by providing an already existing numpy array containing the
        audio samples.

        The shape of the audio samples are always (Nsamples_per_channel, Nchannels).
        """

        self._logger = logging.getLogger(__name__)

        # We sometimes divide by the sample rate to get time values
        assert fs > 0,  "sample rate cannot be zero or negative"

        self.fs         = fs    # sample rate should always be specified in the constructor
        self.nofsamples = None  # number of samples per channel
        self.duration   = None  # duration (length) in seconds
        self.ch         = None  # number of channels
        self._comment   = ''

        if initialdata is None:
            # if we are not given any initial samples we create an empty array of
            # zeros for the audio samples.
            assert isinstance(channels, int)
            assert not(nofsamples!=0 and duration is not None), "choose either samples or duration"

            self.ch = channels

            if duration is not None:
                self.nofsamples = int(duration*self.fs)
                self.duration   = duration
            else:
                self.nofsamples = nofsamples
                self._set_duration()

            # create space for the samples
            self.samples = np.zeros((self.nofsamples, self.ch), dtype=dtype)

        else:
            # An array of initial samples are given, use this to extract
            # channel count and durations.
            assert isinstance(initialdata, np.ndarray), \
                'Only numpy arrays are allowed as initial data'

            assert channels == 0, "parameter 'channels' is redundant if initial data is specified"
            assert nofsamples == 0, "parameter 'nofsamples' is redundant if initial data is specified"
            assert duration is None, "parameter 'duration' is redundant if initial data is specified"

            # copy the data to avoid unexpected data corruption
            self.samples = initialdata.copy()

            if self.samples.ndim == 1:
                # if the array is
                #     array([ 1.,  1.,  1.])
                # we expand it to
                #     array([[ 1.],
                #            [ 1.],
                #            [ 1.]])
                #
                self.samples = np.expand_dims(self.samples, axis=1)

            assert self.samples.ndim == 2, 'shape must be (Nsamples, Nchannels)'

            self.nofsamples, self.ch = self.samples.shape

            # initial data is assumed to have more samples than channels
            assert self.nofsamples > self.ch, 'shape must be (Nsamples, Nchannels)'

            self._set_duration()

        assert self.nofsamples is not None
        assert self.duration   is not None
        assert self.ch         is not None

    def __str__(self):
        s  = '=======================================\n'
        s += 'classname        : %s\n'          %self.__class__.__name__
        s += 'sample rate      : %.1f [Hz]\n'   %self.fs
        s += 'channels         : %i\n'          %self.ch
        s += 'duration         : %.3f [s]\n'    %self.duration
        s += 'datatype         : %s\n'          %self.samples.dtype
        s += 'samples per ch   : %i\n'          %self.nofsamples
        s += 'data size        : %.3f [Mb]\n'   %(self.samples.nbytes/(1024*1024))
        s += 'has comment      : %s\n'          %('yes' if len(self._comment)!=0 else 'no')
        if self.ch != 0:
            # += '-----------------:---------------------\n'
            s += 'peak             : %s\n'  %np.array_str(self.peak()[0],
                                                          precision=4, suppress_small=True)
            s += 'RMS              : %s\n'  %np.array_str(self.rms(),
                                                          precision=4, suppress_small=True)
            s += 'crestfactor      : %s\n'  %np.array_str(self.crest_factor(),
                                                          precision=4, suppress_small=True)
        s += '-----------------:---------------------\n'
        return s

    def __len__(self):
        return self.nofsamples

    def _set_duration(self):
        """internal method

        If we have modified the samples variable (by padding with zeros
        for example) we need to re-calculate the duration
        """
        self.duration = self.nofsamples/self.fs

    def _set_samples(self, idx=0, samples=None):
        """internal method

        NOTE: idx != channel

        idx is always zero indexed since it refers to the numpy array. Channels
        are always indexed from one since this is the natural way of identifying
        channel numbers.
        """
        assert isinstance(samples, np.ndarray)
        assert len(samples) == self.nofsamples
        self.samples[:,idx] = samples

    def pretty_string_samples(self, idx_start=0, idx_end=20, precision=4, header=False):
        s = ''
        if header:
            t = '  '
            u = 'ch'
            for i in range(self.ch):
                t += '-------:'
                u += '  %2i   :' %(i+1)
            t += '\n'
            u += '\n'

            s += t  #   -------:-------:-------:
            s += u  # ch   1   :   2   :   3   :
            s += t  #   -------:-------:-------:

        s += np.array_str(self.samples[idx_start:idx_end,:],
                          max_line_width=260,   # we can print 32 channels before linewrap
                          precision=precision,
                          suppress_small=True)
        if (idx_end-idx_start) < self.nofsamples:
            s  = s[:-1] # strip the right ']' character
            s += '\n ...,\n'
            lastlines = np.array_str(self.samples[-3:,:],
                                     max_line_width=260,
                                     precision=precision,
                                     suppress_small=True)
            s += ' %s\n' %lastlines[1:] # strip first '['
        return s

    def pad(self, nofsamples=0):
        """Zero pad *at the end* of the current audio data.

        increases duration by samples/fs
        """
        assert nofsamples >= 0, "Can't append negative number of samples"
        zeros = np.zeros((nofsamples, self.ch), dtype=self.samples.dtype)
        self.samples = np.append(self.samples, zeros, axis=0)

        self.nofsamples=len(self.samples)
        self._set_duration()

    def trim(self, start=None, end=None):
        """Trim samples **IN PLACE** """
        self.samples = self.samples[start:end]
        self.nofsamples=len(self.samples)
        self._set_duration()

    def _fade(self, millisec, direction):
        """Internal method.

        Fade in/out is essentially the same exept the slope (and position) of the
        ramp. Currently only a linear ramp is implemented.
        """
        assert np.issubdtype(self.samples.dtype, np.floating), \
            "only floating point processing implemented"
        assert millisec >= 0, "Got a time machine?"
        assert direction in ("in", "out")

        fade_seconds = millisec/1000
        assert self.duration > fade_seconds, "fade cannot be longer than the length of the audio"

        sample_count = np.ceil(fade_seconds*self.fs)
        self._logger.debug("fade %s sample count: %i" %(direction, sample_count))

        # generate the ramp
        if direction is "out":
            # ramp down
            ramp = np.linspace(1, 0, num=sample_count, endpoint=True)
        else:
            # ramp up
            ramp = np.linspace(0, 1, num=sample_count, endpoint=True)

        ones = np.ones(len(self)-len(ramp))

        # glue the ones and the ramp together
        if direction is "out":
            gains = np.append(ones, ramp, axis=0)
        else:
            gains = np.append(ramp, ones, axis=0)

        # expand the dimension so we get a one channels array of samples,
        # as in (samples, channels)
        gains = np.expand_dims(gains, axis=1)

        assert len(gains) ==  len(self)

        # repeat the gain vector so we get as many gain channels as all the channels
        gains = np.repeat(gains, self.ch, axis=1)

        assert gains.shape == self.samples.shape

        # apply gains
        self.samples = self.samples * gains

    def fade_in(self, millisec=10):
        """Fade in over 'millisec' seconds. Applies on *all* channels"""
        self._fade(millisec, "in")

    def fade_out(self, millisec=30):
        """Fade out over 'millisec' seconds. Applies on *all* channels"""
        self._fade(millisec, "out")

    def delay(self, n, channel=1):
        """Delay channel x by n samples"""
        self.samples[:,channel-1] = np.pad(self.samples[:,channel-1], (n, 0),
                                           mode="constant")[:-n]

    def get_time(self):
        """Return a vector of time values, starting with t0=0. Useful when plotting."""
        return np.linspace(0, self.duration, num=self.nofsamples, endpoint=False)

    def comment(self, comment=None):
        """Modify or return a string comment."""
        assert isinstance(comment, (str, type(None))), "A comment is a string"

        if comment is not None:
            self._comment = comment

        return self._comment

    def append(self, *args):
        """Add (append) channels *to the right* of the current audio data.

        does zeropadding
        increases channel count
        """
        for i, other in enumerate(args):
            assert isinstance(other, Audio), "only Audio() instances can be used"

            self._logger.debug("** iteration %02i --> appending %s" %((i+1), other.__class__.__name__))
            assert self.fs == other.fs, "Sample rates must match (%s != %s)" %(self.fs, other.fs)
            assert self.samples.dtype == other.samples.dtype, \
                "Data types must match (%s != %s)"%(self.samples.dtype, other.samples.dtype)

            max_nofsamples = max(self.nofsamples,  other.nofsamples)
            missingsamples = abs(self.nofsamples - other.nofsamples)

            self._logger.debug("max nof samples: %i" %max_nofsamples)
            self._logger.debug("appending %i new channel(s) and %i samples" %(other.ch, missingsamples))

            if self.nofsamples > other.nofsamples:
                self._logger.debug("self.nofsamples > other.nofsamples")

                tmp = np.append(other.samples,
                                np.zeros(((missingsamples), other.ch), dtype=other.samples.dtype),
                                axis=0)
                self.samples = np.append(self.samples, tmp, axis=1)

            elif self.nofsamples < other.nofsamples:
                self._logger.debug("self.nofsamples < other.nofsamples")

                tmp = np.append(self.samples,
                                np.zeros(((missingsamples), self.ch), dtype=self.samples.dtype),
                                axis=0)
                self.samples = np.append(tmp, other.samples, axis=1)

            else:
                self._logger.debug("self.nofsamples == other.nofsamples")
                self.samples = np.append(self.samples, other.samples, axis=1)

            self.ch = self.ch+other.ch
            self.nofsamples=max_nofsamples
            self._set_duration()

    def concat(self, *args):
        """Concatenate (append) samples *after* the current audio data.

        example:
            x1 = 1234
            x2 = 5678
            x1.concat(x2) --> 12345678

        """
        for i, other in enumerate(args):
            assert isinstance(other, Audio), "only Audio() instances can be used"

            self._logger.debug("** iteration %02i --> appending %s" %((i+1), other.__class__.__name__))
            assert self.fs == other.fs, "Sample rates must match (%s != %s)" %(self.fs, other.fs)
            assert self.samples.dtype == other.samples.dtype, \
                "Data types must match (%s != %s)"%(self.samples.dtype, other.samples.dtype)
            assert self.ch == other.ch, "channel count must match"

            self.samples = np.append(self.samples, other.samples, axis=0)

            self.nofsamples=len(self.samples)
            self._set_duration()

    def gain(self, *args):
        """Apply gain to the audio samples. Always specify gain values in dB.

        Converts **IN PLACE**
        """
        self._logger.debug('gains: %s' %str(args))

        dt  = self.samples.dtype
        lin = db2lin(args)

        # apply the (linear) gain
        self.samples = lin*self.samples

        # make sure that the data type is retained
        self.samples = self.samples.astype(dt)

    def rms(self):
        """Calculate the RMS (Root Mean Square) value of the audio
        data. Returns the RMS value for each individual channel
        """
        if not (self.samples == 0).all():
            if np.issubdtype(self.samples.dtype, np.floating):
                rms = np.sqrt(np.mean(np.power(self.samples, 2), axis=0))
            else:
                # use a bigger datatype for ints since we most likely will
                # overflow when calculating to the power of 2
                bigger  = np.asarray(self.samples, dtype=np.int64)
                rms     = np.sqrt(np.mean(np.power(bigger, 2), axis=0))

        elif len(self.samples) == 0:
            # no samples are set but channels are configured
            rms = np.zeros(self.ch)
            rms[:] = float('nan')
        else:
            rms = np.zeros(self.ch)
        return rms

    def peak(self):
        """Calculate peak sample value (with sign)"""

        if len(self.samples) != 0:
            if np.issubdtype(self.samples.dtype, np.floating):
                idx = np.absolute(self.samples).argmax(axis=0)
            else:
                # We have to be careful when checking two's complement since the absolute value
                # of the smallest possible value can't be represented without overflowing. For
                # example: signed 16bit has range [-32768, 32767] so abs(-32768) cannot be
                # represented in signed 16 bits --> use a bigger datatype
                bigger  = np.asarray(self.samples, dtype=np.int64)
                idx     = np.absolute(bigger).argmax(axis=0)

            peak = np.array([self.samples[row,col] for col, row in enumerate(idx)])
        else:
            # no samples are set but channels are configured
            idx  = np.zeros(self.ch, dtype=np.int64)
            peak = np.zeros(self.ch)
            peak[:] = float('nan')

        return peak, idx

    def crest_factor(self):
        """Calculate the Crest Factor (peak over RMS) value of the
        audio. Returns the crest factor value for each channel.
        Some common crest factor values:

            sine     : 1.414...
            MLS      : 1    (if no emphasis filter is applied)
            impulse  : very high. The value gets higher the longer
                       the length of the audio data.
            square   : 1    (ideal square)
            zeros    : NaN (we cannot calculate 0/0)

        """
        rms = self.rms()
        assert len(rms) != 0

        with np.errstate(invalid='ignore'):
            # if the rms is zero we will get division errors. Ignore them.
            if len(self.samples) != 0:
                crest = np.abs(self.samples).max(axis=0)/rms
            else:
                # no samples are set but channels are configured
                crest = np.zeros(self.ch)
                crest[:] = float('nan')

        return crest

    def convert_to_integer(self, targetbits=16):
        """Scale floating point values between [-1.0, 1.0] to the equivalent
        signed integer value. Converts **IN PLACE**

        Note: 24 bit signed integers and 8 bit unsigned integers currently unsupported.
        """
        assert targetbits in (8, 16, 32, 64)
        assert self.samples.dtype in (np.int8, np.int16, np.int32, np.int64,
                                      np.float32, np.float64)
        dt = { 8 : 'int8',
              16 : 'int16',
              32 : 'int32',
              64 : 'int64'}

        sourcebits = self.samples.itemsize * 8

        if self.samples.dtype in (np.float32, np.float64):
            self._logger.debug("source is %02i bits (float),   target is %2i bits (integer)"
                              %(sourcebits, targetbits))

            self.samples = np.array(self.samples*(2**(targetbits-1)-1),
                                    dtype=dt.get(targetbits))
        else:
            self._logger.debug("source is %02i bits (integer), target is %2i bits (integer)"
                              %(sourcebits, targetbits))
            raise NotImplementedError("TODO: implement scale int->int")

    def convert_to_float(self, targetbits=64):
        """Scale integer values to equivalent floating point values
        between [-1.0, 1.0]. Converts **IN PLACE**
        """
        assert targetbits in (32, 64)
        assert self.samples.dtype in (np.int8, np.int16, np.int32, np.int64,
                                      np.float32, np.float64)
        dt = {32 : 'float32',
              64 : 'float64'}

        sourcebits = self.samples.itemsize * 8

        if self.samples.dtype in (np.int8, np.int16, np.int32, np.int64):
            self._logger.debug("source is %02i bits (integer), target is %2i bits (float)"
                              %(sourcebits, targetbits))

            self.samples = np.array(self.samples/(2**(sourcebits-1)), dtype=dt.get(targetbits))

        else:
            self._logger.debug("source is %02i bits (float),   target is %2i bits (float)"
                              %(sourcebits, targetbits))

            self.samples = np.array(self.samples, dtype=dt.get(targetbits))

    def write_wav_file(self, filename=None):
        """Save audio data to .wav file."""

        assert filename is not None, "Specify a filename, for example 'filename=audio.wav'"

        self._logger.debug("writing file %s" %filename)
        if self.samples.dtype == np.float64:
            self._logger.warn("datatype is %s" %self.samples.dtype)

        try:
            scipy.io.wavfile.write(filename, self.fs, self.samples)
        except:
            self._logger.exception("Could not write file: '%s'" %filename)

    def plot(self, ch=1, plotname=None, **kwargs):
        """Plot the audio data on a time domain plot.

        example:

            x1 = Sinetone(f0=0.2, fs=10, nofsamples=50)
            x1.plot(linestyle='--', marker='x', color='r', label='sine at 0.2Hz')

        """
        # TODO: add range to plotdata [None:None] is everything

        if ch != 'all':
            assert ch-1 < self.ch, "channel does not exist"

        plt.figure(1)
        plt.title("%s" %self.__class__.__name__)
        if ch != 'all':
            plt.plot(self.get_time(), self.samples[:,ch-1], **kwargs)
        else:
            plt.plot(self.get_time(), self.samples, **kwargs)
        plt.xlabel('Time [s]')
        plt.ylabel('Amplitude [linear]')
        if 'label' in kwargs:
            plt.legend(loc='best')
        plt.grid(True)

        if plotname is None:
            plt.show()
        else:
            plt.savefig(plotname)
            plt.close(1)

    def plot_fft(self, plotname=None, window='hann', normalise=True, **kwargs):
        """Make a plot (in the frequency domain) of all channels"""

        ymin = kwargs.get('ymin', -160) #dB

        freq, mag = self.fft(window=window, normalise=normalise)

        fig_id = 1
        plt.figure(fig_id)

        #plt.semilogx(freq, mag, **kwargs)   # plots all channel directly
        for ch in range(self.ch):
            plt.semilogx(freq, mag[:,ch], label='ch%2i' %(ch+1))

        plt.xlim(left=1)    # we're not interested in freqs. below 1 Hz
        plt.ylim(bottom=ymin)

        plt.xlabel('Frequency [Hz]')
        plt.ylabel('Magnitude [dB]')

        plt.legend(loc='best')
        plt.grid(True)

        if plotname is None:
            plt.show()
        else:
            plt.savefig(plotname)
            plt.close(fig_id)

    def fft(self, window='hann', normalise=True):
        """Calculate the FFT of all channels. Returns data up to fs/2"""
        fftsize = self.nofsamples

        # Avoid Mersenne Primes
        if fftsize in [(2**13)-1, (2**17)-1, (2**19)-1, (2**31)-1]:
            self._logger.warn("FFT size is a Mersenne Prime, increasing size by 1")
            fftsize = fftsize+1

        self._logger.debug("fftsize: %i" %fftsize)
        self._logger.debug("window : %s" %str(window))

        win = scipy.signal.windows.get_window(window, Nx=self.nofsamples) # not fftsize!
        win = np.expand_dims(win, axis=1)
        y   = self.samples*win

        Y   = np.fft.fft(y, n=fftsize, axis=0)
        if normalise:
            Y = Y/fftsize

        mag = lin2db(np.abs(Y))
        frq = np.fft.fftfreq(fftsize, 1/self.fs)

        frq = frq[:int(fftsize/2)]
        mag = mag[:int(fftsize/2)]

        return frq, mag

    def dither(self, bits=16, distribution='TPDF'):
        raise NotImplementedError('TODO')

        assert distribution == 'TPDF', \
            "Only the Triangular Probability Density Function is implemented"

        # Triangular Probability Density Function
        noise = np.random.triangular(-1, 0, 1, self.samples.shape)

    def resample(self, targetrate=44100):
        raise NotImplementedError('TODO')

    def set_sample_rate(self, new_fs):
        """Change the sample rate fs *without* up/down sample conversion.
        This would be the same as changing the playback speed. All data is
        left intact and only the time parameters (fs and duration) are
        changed.
        """
        ratio   = new_fs/self.fs
        self.fs = new_fs

        self._logger.debug('ratio: %.3f' %ratio)
        self._set_duration()
        return ratio

    def normalise(self):
        """Normalise samples so that the new range is
        [-1.0,  1.0] for floats

        Converts **IN PLACE**

        TODO: verify
        [-2^n, 2^n-1] for ints
        """
        peaks, unused_idx = self.peak()
        self._logger.debug("raw peaks: %s" %peaks)

        max_abs = np.max(np.absolute(peaks))
        self._logger.debug("max_abs: %s" %max_abs)

        self.samples = self.samples/max_abs

        peaks, unused_idx = self.peak()
        self._logger.debug("new peaks: %s" %peaks)

#===================================================================================================
# Audio sub-classes
#===================================================================================================

class Sinetone(Audio):
    def __init__(self, f0=997, fs=96000, duration=None, gaindb=0, nofsamples=0, phasedeg=0):
        """Generate a sine tone"""
        assert f0 < fs/2, "Sampling theorem is violated"
        Audio.__init__(self, channels=1, fs=fs, nofsamples=nofsamples, duration=duration)

        self.f0         = f0
        self.phasedeg   = phasedeg

        self._set_samples(idx=0, samples=self._sine_gen(f0, phasedeg))
        self.gain(gaindb)

    def _sine_gen(self, freq, pha):
        return np.sin(2*np.pi*freq*self.get_time()+np.deg2rad(pha))

    def __repr__(self):
        assert self.ch == 1, "If a channel has been appended we don't know anything about its data"

        s = 'Sinetone(f0=%r, fs=%r, nofsamples=%r, gaindb=%r, phasedeg=%r)' \
            %(self.f0, self.fs, self.nofsamples,
              lin2db(abs(float(self.peak()[0]))),  # only one channel here.
              self.phasedeg)
        return s

    def __str__(self):
        s  = Audio.__str__(self)
        s += 'frequency        : %.1f [Hz]\n'   %self.f0
        s += 'phase            : %.1f [deg]\n'  %self.phasedeg
        s += '-----------------:---------------------\n'
        return s

    def set_sample_rate(self, new_fs):
        ratio = Audio.set_sample_rate(self, new_fs)
        self.f0 = ratio * self.f0

class Sinetones(Sinetone):
    def __init__(self, *args, **kwargs):
        """Generate multiple sinetones. This is a quick way to generate multichannel audio.
        Each positional argument generates a sinetone at that channel. Setting the frequency
        to 0 guarantees that the channel is muted (contains samples with the value 0).
        Keywords accepted are similar to the ones used in the Sinetone() class.

        Example:

            >>> x = Sinetones(200, 500, 900, fs=24000, duration=1.5, gaindb=-6, phasedeg=0)
            >>> print(x)
            =======================================
            classname        : Sinetones
            sample rate      : 24000.0 [Hz]
            channels         : 3
            duration         : 1.500 [s]
            datatype         : float64
            samples per ch   : 36000
            data size        : 0.824 [Mb]
            has comment      : no
            peak             : [ 0.5012  0.5012 -0.5012]
            RMS              : [ 0.3544  0.3544  0.3544]
            crestfactor      : [ 1.4142  1.4142  1.4142]
            -----------------:---------------------
            phase (all ch)   : 0.0 [deg]
                             :
            channel  1       : 200.0 [Hz]
            channel  2       : 500.0 [Hz]
            channel  3       : 900.0 [Hz]
            -----------------:---------------------
            >>>

        The gaindb argument can be an iterable of the same length as the number of frequencies
        specified. In this case a gain can be applied individually for each channel.

            >>> x = Sinetones(1000, 2000, duration=1, gaindb=(-6, -20))

        A list can be used as the argument for the frequencies. Use the * notation to expand
        the list:

            >>> import numpy as np
            >>> f = np.zeros(8)
            >>> f[3] = 700
            >>> f[7] = 2000
            >>> x = Sinetones(*f, duration=1)
            >>> print(x)
            =======================================
            classname        : Sinetones
            sample rate      : 96000.0 [Hz]
            channels         : 8
            duration         : 1.000 [s]
            datatype         : float64
            samples per ch   : 96000
            data size        : 5.859 [Mb]
            has comment      : no
            peak             : [ 0.  0.  0. -1.  0.  0.  0.  1.]
            RMS              : [ 0.      0.      0.      0.7071  0.      0.      0.      0.7071]
            crestfactor      : [    nan     nan     nan  1.4142     nan     nan     nan  1.4142]
            -----------------:---------------------
            phase (all ch)   : 0.0 [deg]
                             :
            channel  1       :
            channel  2       :
            channel  3       :
            channel  4       : 700.0 [Hz]
            channel  5       :
            channel  6       :
            channel  7       :
            channel  8       : 2000.0 [Hz]
            -----------------:---------------------
            >>>

        The argument phasedeg applies to all channels.
        """

        fs                  = kwargs.pop('fs',          96000)
        duration            = kwargs.pop('duration',    None)
        nofsamples          = kwargs.pop('nofsamples',  0)
        self._gaindb        = kwargs.pop('gaindb',      0)
        self.phasedeg       = kwargs.pop('phasedeg',    0)
        self.frequencies    = args

        for frequency in self.frequencies:
            assert frequency < fs/2, "Sampling theorem is violated for frequency %.1f" %frequency

        if not isinstance(self._gaindb, int):
            assert len(self._gaindb) == len(self.frequencies), \
                "set as many gains as channels used: %i != %i" %(len(self._gaindb),
                                                                 len(self.frequencies))

        Audio.__init__(self, channels=len(self.frequencies), fs=fs, nofsamples=nofsamples,
                       duration=duration)

        for i, frequency in enumerate(self.frequencies):
            if frequency != 0:
                self._set_samples(idx=i, samples=self._sine_gen(frequency, self.phasedeg))
            else:
                pass # channel is silence

        self.gain(self._gaindb)

    def __repr__(self):
        s = 'Sinetones(*%r, fs=%r, nofsamples=%r, gaindb=%r, phasedeg=%r)' \
            %(list(self.frequencies), self.fs, self.nofsamples,self._gaindb,self.phasedeg)
        return s

    def __str__(self):
        s  = Audio.__str__(self)
        s += 'phase (all ch)   : %.1f [deg]\n'  %self.phasedeg
        s += '                 :\n'
        for i, frequency in enumerate(self.frequencies):
            if frequency != 0:
                s += 'channel %2i       : %.1f [Hz]\n'   %(i+1, frequency)
            else:
                s += 'channel %2i       :\n'   %(i+1)
        s += '-----------------:---------------------\n'
        return s

    def set_sample_rate(self, new_fs):
        ratio = Audio.set_sample_rate(self, new_fs)
        self.frequencies = [ratio*f for f in self.frequencies]

class SquareWave(Audio):
    def __init__(self, f0=997, fs=96000, duration=None, gaindb=0, nofsamples=0,
                 phasedeg=0, dutycycle=0.5):
        """Generate an ideal squarewave."""
        assert f0 < fs/2, "Sampling theorem is violated"
        assert dutycycle < 1 and dutycycle > 0
        Audio.__init__(self, channels=1, fs=fs, nofsamples=nofsamples, duration=duration)

        self.f0         = f0
        self.phasedeg   = phasedeg
        self.dutycycle  = dutycycle

        samples = scipy.signal.square(2*np.pi*f0*self.get_time()+np.deg2rad(phasedeg),
                                      duty=dutycycle)
        self._set_samples(idx=0, samples=samples)
        self.gain(gaindb)

    def __repr__(self):
        assert self.ch == 1, "If a channel has been appended we don't know anything about its data"

        s = 'SquareWave(f0=%r, fs=%r, gaindb=%r, nofsamples=%r, phasedeg=%r, dutycycle=%r)' \
            %(self.f0, self.fs,
              lin2db(abs(float(self.peak()[0]))), # only one channel here.
              self.nofsamples, self.phasedeg, self.dutycycle)
        return s

    def __str__(self):
        s  = Audio.__str__(self)
        s += 'frequency        : %.1f [Hz]\n'       %self.f0
        s += 'phase            : %.1f [deg]\n'      %self.phasedeg
        s += 'duty cycle       : %.3f (%4.1f%%)\n'  %(self.dutycycle, self.dutycycle*100)
        s += '-----------------:---------------------\n'
        return s

    def set_sample_rate(self, new_fs):
        ratio = Audio.set_sample_rate(self, new_fs)
        self.f0 = ratio * self.f0

class FourierSeries(Sinetone):
    def __init__(self, f0=997, fs=96000, duration=None, gaindb=0, nofsamples=0,
                 phasedeg=0, harmonics=7,):
        """Construct a square wave by adding odd harmonics with decreasing
        amplitude, i.e. Fourier Series.
        """
        Sinetone.__init__(self, f0=f0, phasedeg=phasedeg, fs=fs, nofsamples=nofsamples,
                          duration=duration, gaindb=0)

        assert harmonics >= 0

        self.harmonics = harmonics
        self._logger.debug("fundamental f0: %.1f" %f0)

        for n in range(3, 2*(self.harmonics+1), 2):
            if n <= 15:
                self._logger.debug("adding harmonic n: %2i with amplitude 1/%i" %(n, n))
            if n == 17:
                self._logger.debug("adding %i more harmonics..." %(self.harmonics-(n-3)//2))

            #self.samples[:,0] += np.sin(2*np.pi*(n*f0)*self.get_time()+np.deg2rad(phasedeg*n))/n
            self.samples[:,0] += (1/n)*self._sine_gen(n*f0, n*phasedeg)
        self.gain(gaindb)

    def __repr__(self):
        assert self.ch == 1, "If a channel has been appended we don't know anything about its data"

        s = 'FourierSeries(f0=%r, fs=%r, gaindb=%r, nofsamples=%r, phasedeg=%r, harmonics=%r)' \
            %(self.f0, self.fs,
              lin2db(abs(float(self.peak()[0]))), # only one channel here.
              self.nofsamples, self.phasedeg, self.harmonics)
        return s

    def __str__(self):
        s  = Sinetone.__str__(self)
        s = s.rstrip('-----------------:---------------------\n')
        s += '\n'
        s += 'harmonics        : %i \n' %self.harmonics
        s += '-----------------:---------------------\n'
        return s

class Noise(Audio):
    colours = ('white', 'pink', 'brown', 'blue', 'violet', 'grey')

    def __init__(self, channels=1, fs=96000, duration=None, gaindb=-10, nofsamples=0,
                 colour='white'):
        """Generate uncorrelated noise.

        white       : flat power spectral density
        pink        : -3dB per octave
        brown(ian)  : -6dB per octave
        blue        : +3dB per octave
        violet      : +6dB per octave
        grey        : equal loudness
        """

        assert colour in Noise.colours, "choose the colour of the noise: %s" %str(Noise.colours)
        Audio.__init__(self, channels=channels, fs=fs, nofsamples=nofsamples, duration=duration)
        # the distribution in np.random.uniform is half open, i.e -1.0 is
        # included but 1.0 is not. Possible fix: use integers instead, then
        # scale to floats. Might not work, since the integers will be
        # represented using twos complement and we then have an asymmetrical
        # range anyhow.

        self._colour = colour

        # first generate uniformly distributed noise, i.e. white noise. Then filter
        # to get the required shape.
        for ch in range(channels):
            self._set_samples(idx=ch,
                              samples=np.random.uniform(low=-1.0, high=1.0, size=self.nofsamples))
        if self._colour=='pink':
            # -3dB per octave
            self._logger.debug("filtering to get pink noise")
            # http://dsp.stackexchange.com/q/322/6999
            B = [0.049922035, -0.095993537, 0.050612699, -0.004408786]
            A = [1, -2.494956002, 2.017265875, -0.522189400]
            self.samples = scipy.signal.lfilter(B, A, self.samples, axis=0)

        elif self._colour=='brown':
            # -6dB per octave
            raise NotImplementedError('TODO')

        elif self._colour=='blue':
            # +3dB per octave
            raise NotImplementedError('TODO')

        elif self._colour=='violet':
            # +6dB per octave
            raise NotImplementedError('TODO')

        elif self._colour=='grey':
            # equal loudness
            raise NotImplementedError('TODO')

        self.gain(gaindb)

    def __str__(self):
        s  = Audio.__str__(self)
        s += 'colour           : %s\n'  %self._colour
        s += '-----------------:---------------------\n'
        return s

class WavFile(Audio):
    def __init__(self, filename=None, scale2float=True):
        """Read a .wav file from disk"""
        assert filename is not None, "Specify a filename"
        self.filename = filename

        fs, samples = scipy.io.wavfile.read(filename)
        if samples.ndim == 1:
            samples = np.expand_dims(samples, axis=1)

        Audio.__init__(self, fs=fs, initialdata=samples)

        del samples # just to make sure

        if scale2float:
            self.convert_to_float(targetbits=64)

    def __str__(self):
        s  = Audio.__str__(self)
        s += 'filename         : %s\n'  %os.path.basename(self.filename)
        s += '-----------------:---------------------\n'
        return s

#===================================================================================================
# Functions
#===================================================================================================

def lin2db(lin):
    with np.errstate(divide='ignore'):
        # linear value 0 is common (as -inf dB) so we ignore any division warnings
        db = 20*np.log10(lin)
    return db

def pow2db(power):
    with np.errstate(divide='ignore'):
        # ignore any division warnings
        db = 10*np.log10(power)
    return db

def db2lin(db):
    lin = np.power(10, np.array(db)/20)
    return lin

def db2pow(db):
    power = np.power(10, np.array(db)/10)
    return power

def speed_of_sound(temperature=20, medium='air'):
    """The speed of sound is depending on the medium and the temperature. For air at
    a temperature of 20 degree Celcius the speed of sound is approximately 343 [m/s]
    """
    assert medium in ['air',], "TODO: water, iron"

    c = float('nan')

    if medium == 'air':
        c = 331.3*np.sqrt(1+temperature/273.15)

    return c

def wavelength(frequency, speed=343.2):
    """Calculate the wavelength l of frequency f given the speed (of sound)"""
    l = speed/frequency
    return l

def rad2hz(w0, fs=96000):
    """Calculate a normalised rotational frequency so that w0=2*pi --> f0=fs

                    w0
        f0 = fs * ------
                   2*pi
    """
    return fs*np.array(w0)/(2*np.pi)

def hz2rad(f0, fs=96000):
    """Calculate a normalised angular frequency so that f0=fs --> w0=2*pi

               1
        w0 = ----- * 2*pi*f0
              fs
    """
    return (1/fs)*2*np.pi*np.array(f0)

__all__ = [
           # classes
           'Audio',
           'Sinetone',
           'Sinetones',
           'SquareWave',
           'FourierSeries',
           'Noise',
           'WavFile',

           # functions
           'lin2db',
           'pow2db',
           'db2lin',
           'db2pow',
           'speed_of_sound',
           'wavelength',
           'rad2hz',
           'hz2rad',
           ]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')

    print('-- Done --')



"""
Music package

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2016 Ronny Andersson
@license: MIT
"""

from . import scales
from . import spn

__all__ = [
           'scales',
           'spn',
           ]



"""
Created on 16 Feb 2014

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2014 Ronny Andersson
@license: MIT
"""

# standard library
import logging
from abc import ABCMeta, abstractmethod

# external libraries
import numpy as np

# local libraries
from zignal.filters.linearfilter import IIR

class Biquad(IIR):
    """Implements a two-pole, two-zeros biquadratic filter, biquad."""
    def __init__(self, B=None, A=None, fs=96000):
        IIR.__init__(self, B=B, A=A, fs=fs)

    def set_coefficients(self, B=None, A=None):
        """Overrides the method in the parent class so we can assure
        that we always have three B and three A coefficients.
        """
        self._logger.debug("set coefficients (class Biquad)")

        self._B = np.array((1,0,0)) if B is None else np.array(B)
        self._A = np.array((1,0,0)) if A is None else np.array(A)

        assert len(self._B) == 3, "Biquads have three B coefficients"
        assert len(self._A) == 3, "Biquads have three A coefficients"

class BiquadNormalised(Biquad):
    """The normalised biquad always has A[0]=1.0

    This particular implementation keeps the number of coefficients to six in total,
    making sure that the normalise() method is called whenever we calculate the
    coefficients.
    """
    def __init__(self, B=None, A=None, fs=96000):
        Biquad.__init__(self, B=B, A=A, fs=fs)
        self.normalise()

    def set_coefficients(self, B=None, A=None):
        self._logger.debug("set coefficients (class BiquadNormalised)")
        Biquad.set_coefficients(self, B=B, A=A)
        self.normalise()

    def get_coefficients_Pd(self):
        """Return coefficients compatible with the [biquad~] object in Pd"""
        B, A = self.get_coefficients()
        return (B[0], B[1], B[2], -A[1], -A[2]) #FIXME: Verify this

    def get_coefficients_MaxMSP(self):
        """Return coefficients compatible with the [biquad~] object in Max/MSP"""
        B, A = self.get_coefficients()
        return (B[0], B[1], B[2], A[1], A[2])   #FIXME: Verify this

class _BiquadParametric(BiquadNormalised, metaclass=ABCMeta):
    def __init__(self, filtertype=None, gaindb=0, f0=997, Q=0.707, fs=96000):
        BiquadNormalised.__init__(self, B=None, A=None, fs=fs)
        self._verify_parameters(filtertype, gaindb, f0, Q)
        self.calculate_coefficients(filtertype=filtertype, gaindb=gaindb, f0=f0, Q=Q)

    def _verify_parameters(self, filtertype, gaindb, f0, Q):
        """Internal verification that we are at least partially sane before
        proceeding.
        """
        assert filtertype is not None, "Specify a filter type (lowpass, highpass, peak, ..."
        assert f0 >= 0,             "negative frequency is not allowed"
        assert f0 < self.fs/2,      "f0 must be below the Nyquist frequency (fs/2)"
        assert Q > 0,               "Q needs to be positive and above zero (we divide by Q)"

        self.filtertype = filtertype
        self.gaindb     = gaindb
        self.f0         = f0
        self.Q          = Q

    def __str__(self):
        s  = BiquadNormalised.__str__(self)
        # += '-----------------:---------------------\n'
        s += 'type             : %s\n'          %self.filtertype
        s += 'gain             : %.2f [dB]\n'   %self.gaindb
        s += 'f0               : %.1f [Hz]\n'   %self.f0
        s += 'Q                : %.4f\n'        %self.Q
        return s

    @abstractmethod
    def calculate_coefficients(self, filtertype=None, gaindb=None, f0=None, Q=None):
        # Implement this method in child class
        pass

class RBJ(_BiquadParametric):
    """An implementation of the Audio EQ cookbook by Robert Bristow-Johnson

    Robert Bristow-Johnson - Cookbook formulae for audio EQ biquad filter coefficients
    http://www.musicdsp.org/files/Audio-EQ-Cookbook.txt
    """

    class Types(object):
        lowpass     = 'lowpass'
        highpass    = 'highpass'
        bandpass1   = 'bandpass1'
        bandpass2   = 'bandpass2'
        notch       = 'notch'
        allpass     = 'allpass'
        peak        = 'peak'
        lowshelf    = 'lowshelf'
        highshelf   = 'highshelf'

    def __init__(self, filtertype=None, gaindb=0, f0=997, Q=0.707, fs=96000):
        _BiquadParametric.__init__(self, filtertype=filtertype, gaindb=gaindb, f0=f0, Q=Q, fs=fs)

    def calculate_coefficients(self, filtertype=None, gaindb=None, f0=None, Q=None):
        self._verify_parameters(filtertype, gaindb, f0, Q)

        # intermediate variables
        _A      = 10**(gaindb/40)
        _w0     = 2 * np.pi * f0/self.fs
        _cos_w0 = np.cos(_w0)
        _sin_w0 = np.sin(_w0)
        _alpha  = _sin_w0/(2*Q)

        if filtertype == self.Types.lowpass:
            # LPF:        H(s) = 1 / (s^2 + s/Q + 1)
            b0 =  (1 - _cos_w0)/2
            b1 =   1 - _cos_w0
            b2 =  (1 - _cos_w0)/2
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.highpass:
            # HPF:        H(s) = s^2 / (s^2 + s/Q + 1)
            b0 =  (1 + _cos_w0)/2
            b1 = -(1 + _cos_w0)
            b2 =  (1 + _cos_w0)/2
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.bandpass1:
            # BPF:        H(s) = s / (s^2 + s/Q + 1)  (constant skirt gain, peak gain = Q)
            b0 =   _sin_w0/2    #  =   Q*_alpha
            b1 =   0
            b2 =  -_sin_w0/2    # =  -Q*_alpha
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.bandpass2:
            # BPF:        H(s) = (s/Q) / (s^2 + s/Q + 1)      (constant 0 dB peak gain)
            b0 =   _alpha
            b1 =   0
            b2 =  -_alpha
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.notch:
            # notch:      H(s) = (s^2 + 1) / (s^2 + s/Q + 1)
            b0 =   1
            b1 =  -2 * _cos_w0
            b2 =   1
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.allpass:
            # APF:        H(s) = (s^2 - s/Q + 1) / (s^2 + s/Q + 1)
            b0 =   1 - _alpha
            b1 =  -2 * _cos_w0
            b2 =   1 + _alpha
            a0 =   1 + _alpha
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha

        elif filtertype == self.Types.peak:
            # peakingEQ:  H(s) = (s^2 + s*(A/Q) + 1) / (s^2 + s/(A*Q) + 1)
            b0 =   1 + _alpha*_A
            b1 =  -2 * _cos_w0
            b2 =   1 - _alpha*_A
            a0 =   1 + _alpha/_A
            a1 =  -2 * _cos_w0
            a2 =   1 - _alpha/_A

        elif filtertype == self.Types.lowshelf:
            # lowShelf: H(s) = A * (s^2 + (sqrt(A)/Q)*s + A)/(A*s^2 + (sqrt(A)/Q)*s + 1)
            b0 =    _A*( (_A+1) - (_A-1)*_cos_w0 + 2*np.sqrt(_A)*_alpha )
            b1 =  2*_A*( (_A-1) - (_A+1)*_cos_w0                        )
            b2 =    _A*( (_A+1) - (_A-1)*_cos_w0 - 2*np.sqrt(_A)*_alpha )
            a0 =         (_A+1) + (_A-1)*_cos_w0 + 2*np.sqrt(_A)*_alpha
            a1 =    -2*( (_A-1) + (_A+1)*_cos_w0                        )
            a2 =         (_A+1) + (_A-1)*_cos_w0 - 2*np.sqrt(_A)*_alpha

        elif filtertype == self.Types.highshelf:
            # highShelf: H(s) = A * (A*s^2 + (sqrt(A)/Q)*s + 1)/(s^2 + (sqrt(A)/Q)*s + A)
            b0 =    _A*( (_A+1) + (_A-1)*_cos_w0 + 2*np.sqrt(_A)*_alpha )
            b1 = -2*_A*( (_A-1) + (_A+1)*_cos_w0                        )
            b2 =    _A*( (_A+1) + (_A-1)*_cos_w0 - 2*np.sqrt(_A)*_alpha )
            a0 =         (_A+1) - (_A-1)*_cos_w0 + 2*np.sqrt(_A)*_alpha
            a1 =     2*( (_A-1) - (_A+1)*_cos_w0                        )
            a2 =         (_A+1) - (_A-1)*_cos_w0 - 2*np.sqrt(_A)*_alpha

        else:
            valid = [i for i in vars(self.Types) if not i.startswith("__")]
            raise NotImplementedError("Valid types are: %s" %valid)

        self.set_coefficients(B=(b0, b1, b2), A=(a0, a1, a2))

class Zolzer(_BiquadParametric):
    """An implementation of Equalizer filters from DAFX - Zolzer et al."""
    class Types(object):
        lowpass     = 'lowpass'
        highpass    = 'highpass'
        peak        = 'peak'
        lowshelf    = 'lowshelf'
        highshelf   = 'highshelf'

    def __init__(self, filtertype=None, gaindb=0, f0=997, Q=0.707, fs=96000):
        _BiquadParametric.__init__(self, filtertype=filtertype, gaindb=gaindb, f0=f0, Q=Q, fs=fs)

    def calculate_coefficients(self, filtertype=None, gaindb=None, f0=None, Q=None):
        self._verify_parameters(filtertype, gaindb, f0, Q)

        K = np.tan(np.pi*f0/self.fs)

        #------------
        # peak
        #------------
        if filtertype == self.Types.peak:
            if gaindb > 0:
                self._logger.debug('peak boost')
                V0  = 10**(gaindb/20)
                den =  1 +  1/Q * K + K**2

                b0  = (1 + V0/Q * K + K**2)         / den
                b1  = (2 * (K**2 - 1))              / den
                b2  = (1 - V0/Q * K + K**2)         / den
                a1  = b1
                a2  = (1 -  1/Q * K + K**2)         / den
            else:
                self._logger.debug('peak cut')
                V0  = 10**(-gaindb/20)
                den =  1 + V0/Q * K + K**2
                b0  = (1 +  1/Q * K + K**2)         / den
                b1  = (          2 * (K**2 - 1))    / den
                b2  = (1 -  1/Q * K + K**2)         / den
                a1  = b1
                a2  = (1 - V0/Q * K + K**2)         / den

        #------------
        # low shelf
        #------------
        elif filtertype == self.Types.lowshelf:
            # Parameter 'Q' is not used
            if gaindb > 0:
                self._logger.debug('lowshelf boost')
                V0  = 10**(gaindb/20)
                den =  1+np.sqrt(2)*K + K**2
                b0  = (1+np.sqrt(V0*2)*K + V0*K**2) / den
                b1  = (2*(V0*K**2-1))               / den
                b2  = (1-np.sqrt(V0*2)*K + V0*K**2) / den
                a1  = (2*(K**2-1))                  / den
                a2  = (1-np.sqrt(2)*K + K**2)       / den
            else:
                self._logger.debug('lowshelf cut')
                V0  = 10**(-gaindb/20)
                den =  1+np.sqrt(2*V0)*K + V0*K**2
                b0  = (1+np.sqrt(2)*K + K**2)       / den
                b1  = (2*(K**2-1))                  / den
                b2  = (1-np.sqrt(2)*K + K**2)       / den
                a1  = (2*(V0*K**2-1))               / den
                a2  = (1-np.sqrt(2*V0)*K + V0*K**2) / den

        #------------
        # high shelf
        #------------
        elif filtertype == self.Types.highshelf:
            # Parameter 'Q' is not used
            if gaindb > 0:
                self._logger.debug('highshelf boost')
                V0  = 10**(gaindb/20)
                den =  1+np.sqrt(2)*K + K**2
                b0  = (V0+np.sqrt(V0*2)*K + K**2)   / den
                b1  = (2*(K**2-V0))                 / den
                b2  = (V0-np.sqrt(V0*2)*K + K**2)   / den
                a1  = (2*(K**2-1))                  / den
                a2  = (1-np.sqrt(2)*K + K**2)       / den
            else:
                self._logger.debug('highshelf cut')
                V0  = 10**(-gaindb/20)
                b0  = (1+np.sqrt(2)*K + K**2)           / (V0+np.sqrt(2*V0)*K +  K**2)
                b1  = (2*(K**2-1))                      / (V0+np.sqrt(2*V0)*K +  K**2)
                b2  = (1-np.sqrt(2)*K + K**2)           / (V0+np.sqrt(2*V0)*K +  K**2)
                a1  = (2*(((K**2)/V0)-1))               / ( 1+np.sqrt(2/V0)*K + (K**2)/V0)
                a2  = (1-np.sqrt(2/V0)*K + (K**2)/V0)   / ( 1+np.sqrt(2/V0)*K + (K**2)/V0)

        #------------
        # low pass
        #------------
        elif filtertype == self.Types.lowpass:
            # Parameter 'Q' is not used
            self._logger.debug('lowpass')

            den =  1+np.sqrt(2)*K + K**2
            b0  = (K**2)                    / den
            b1  = (2*K**2)                  / den
            b2  = (K**2)                    / den
            a1  = (2*(K**2-1))              / den
            a2  = (1-np.sqrt(2)*K + K**2)   / den

        #------------
        # high pass
        #------------
        elif filtertype == self.Types.highpass:
            # Parameter 'Q' is not used
            self._logger.debug('highpass')

            den =  1+np.sqrt(2)*K + K**2
            b0  = (1)                       / den
            b1  = (-2)                      / den
            b2  = (1)                       / den
            a1  = (2*(K**2-1))              / den
            a2  = (1-np.sqrt(2)*K + K**2)   / den

        else:
            raise NotImplementedError()

        self.set_coefficients(B=(b0, b1, b2), A=(1.0, a1, a2))

__all__ = ['Biquad', 'RBJ', 'Zolzer',]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')

    print('++ End of script ++')



'''
Created on 21 Feb 2015

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2015 Ronny Andersson
@license: MIT
'''

# standard library
import logging

# external libraries
import numpy as np

# local libraries
from . import spn

def equal_temperament(n):
    """Twelve-tone equal temperament (12TET) divides the octave into 12
    equal parts, making the interval between two ajacent notes the twelfth
    root of two.

    The argument n can be a number or a list/tuple/iterable.

        2^(1/12):1 --> 2^(n/12)

    https://en.wikipedia.org/wiki/Equal_temperament
    """

    ratio = np.power(2, (np.array(n)/12))

    return ratio

def piano_key2freq(n, a=49, tuning=440):
    """Twelve-tone equal temperament tuning for a theoretically ideal piano.

    The argument n can be a number or a list/tuple/iterable.

    The 49th key, called A4 is tuned to the reference (tuning) frequency, normally
    440Hz. The frequency is then given by

        f(n) = 440*2^((n-49)/12)

    https://en.wikipedia.org/wiki/Piano_key_frequencies
    """

    frequency = tuning*equal_temperament(np.array(n)-a)

    return frequency

def piano_freq2key(f, a=49, tuning=440, quantise=False):
    """Frequency [f] to twelve-tone equal temperament tuning for a theoretically
    ideal piano, where 440Hz-->49
    """

    key = 12*np.log2(f/tuning) + a

    if quantise:
        key = np.int(np.round(key))

    return key

def piano_note2freq(note, tuning=440):
    """Convert a piano note like 'C4' to 12TET frequency 261.6 Hz"""
    idx  = spn.key2index(note)
    freq = piano_key2freq(idx, tuning=tuning)

    return freq

def piano_freq2note(f, tuning=440):
    """Given frequency f, calculate the nearest note in 12TET SPN notation"""
    idx = piano_freq2key(f, tuning=tuning, quantise=True)
    key = spn.index2key(idx)

    return key

def midi_key2freq(n, tuning=440):
    """MIDI Tuning Standard. Convert midi note n to frequency f [Hz]. MIDI note 69
    equals 440 Hz, equal temperament.

    http://en.wikipedia.org/wiki/MIDI_Tuning_Standard
    """

    frequency = piano_key2freq(n, a=69, tuning=tuning)

    return frequency

def midi_freq2key(f, tuning=440, quantise=False):
    """ MIDI Tuning Standard. Convert frequency f [Hz] to a midi note. MIDI note 69
    equals 440 Hz, equal temperament

    69+12*log2(f/440)

    http://en.wikipedia.org/wiki/MIDI_Tuning_Standard
    """

    midinote = piano_freq2key(f, a=69, tuning=tuning, quantise=quantise)

    return midinote

__all__ = [
           'equal_temperament',
           'piano_key2freq',
           'piano_freq2key',
           'piano_note2freq',
           'piano_freq2note',
           'midi_key2freq',
           'midi_freq2key',
           ]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')

    print('-- Done --')



"""
Filters package

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2014 Ronny Andersson
@license: MIT
"""

from . import linearfilter
from . import biquads

__all__ = [
           'biquads',
           'linearfilter',
           ]



from setuptools import setup, find_packages
import os
import ast

# Dynamically calculate the version
version = None
with open(os.path.join('zignal', '__init__.py'), 'r') as fp:
    for line in fp:
        if line.startswith('__version__'):
            version = ast.parse(line).body[0].value.s
            break

assert version is not None, "Problem reading the version from __init__.py"

setup(
    name                = 'zignal',
    author              = 'Ronny Andersson',
    author_email        = 'ronny@andersson.tk',
    packages            = find_packages(),
    install_requires    = [
                           'numpy',
                           'scipy',
                           'matplotlib',
                           ],
    extras_require      = {
                           'sndcard': ['pyaudio'],
                           'testing': ['nose'],
                          },
    url                 = 'https://github.com/ronnyandersson/zignal',
    download_url        = 'https://pypi.python.org/pypi/zignal',
    license             = open('LICENSE.txt').read(),
    version             = version,
    description         = 'Audio signal processing library',
    long_description    = open('README.md').read(),
    platforms           = ['any'],
    entry_points        = {
                           'console_scripts': \
                           [
                            'zignal-listsndcards=zignal.sndcard:list_devices',
                            ],
                           },
    keywords            = [
                           'audio',
                           'sound',
                           'card',
                           'soundcard',
                           'pyaudio',
                           'portaudio',
                           'playback',
                           'recording',
                           'digital',
                           'signal',
                           'processing',
                           'DSP',
                           'signalprocessing',
                           'fourier',
                           'FFT',
                           'filter',
                           'filtering',
                           'parametric',
                           'eq',
                           'equaliser',
                           'equalizer',
                           'biquad',
                           'cookbook',
                           'sine',
                           'generator',
                           'mls',
                           'mlssa',
                           'maximum',
                           'length',
                           'sequence',
                           'maximumlengthsequence',
                           'pseudo',
                           'random',
                           'pseudorandom',
                           'measure',
                           'measurement',
                           'impulse',
                           'response',
                           'impulseresponse',
                           'frequency',
                           'frequencyresponse',
                           'magnitude',
                           'magnituderesponse',
                           'piano',
                           'midi',
                           'tuning',
                           'scale',
                           'pitch',
                           'notation',
                           'equal',
                           'temperament',
                           '12TET',
                           'spn',
                           ],
    classifiers         = [
                           'Development Status :: 3 - Alpha',
                           'Environment :: Console',
                           'Intended Audience :: Developers',
                           'Intended Audience :: Education',
                           'Intended Audience :: Science/Research',
                           'License :: OSI Approved :: MIT License',
                           'Operating System :: OS Independent',
                           'Programming Language :: Python',
                           'Programming Language :: Python :: 3.6',
                           'Topic :: Education',
                           'Topic :: Multimedia :: Sound/Audio :: Analysis',
                           'Topic :: Multimedia :: Sound/Audio :: Capture/Recording',
                           'Topic :: Multimedia :: Sound/Audio :: Editors',
                           'Topic :: Multimedia :: Sound/Audio :: MIDI',
                           'Topic :: Multimedia :: Sound/Audio :: Mixers',
                           'Topic :: Multimedia :: Sound/Audio :: Players',
                           'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis',
                           'Topic :: Scientific/Engineering :: Mathematics',
                           'Topic :: Software Development :: Quality Assurance',
                           'Topic :: Software Development :: Testing',
                           ],
)



'''
Created on 21 Feb 2015

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2015 Ronny Andersson
@license: MIT

Scientific pitch notation (also known as American Standard Pitch Notation
(ASPN) and International Pitch Notation (IPN))
'''

# standard library
import logging
import re

def key2index(key='A4'):
    """ Scientific pitch notation key mapped to an index

    Example:

       >>> key2index('A4')
       49

    A0 is 1, C1 is 4, A4 is 49

    Double sharps are notated with 'x'

    https://en.wikipedia.org/wiki/Scientific_pitch_notation
    https://en.wikipedia.org/wiki/Sharp_(music)
    https://en.wikipedia.org/wiki/Flat_(music)
    """

    logger = logging.getLogger(__name__)

    valid_chars = "ABCDEFGb#x-0123456789"

    for c in key:
        assert c in valid_chars, "Valid characters are: '%s'" %valid_chars

    m = re.match(r"(?P<note>^[A-G])(?P<half>b{0,3}|#{0,1}x{0,1})(?P<octave>[\-]?[0-9]+)", key)
    if m is None:
        raise ValueError("Failed to match key '%s'" %key)

    matched = m.groupdict()

    logger.debug("key: %s re: %s" %(key.ljust(4), matched))

    octave  = int(matched.get('octave'))
    note    = matched.get('note')
    half    = matched.get('half')

    aug_or_dim = {
                  'bbb' : -3,
                  'bb'  : -2,
                  'b'   : -1,
                  '#'   :  1,
                  'x'   :  2,
                  '#x'  :  3,
                  }

    octave  = (octave-1)*12
    note    = " C D EF G A B".index(note)
    #          ^ ^ ^  ^ ^ ^ the spaces are important, we can 'index-search' the array
    half    = aug_or_dim.get(half, 0)

    # A0      = 1
    # A#0/Bb0 = 2
    # B0      = 3
    # C1      = 4
    pitch = 3 + octave + note + half

    logger.debug("key: %s spn: %i" %(key.ljust(4), pitch))

    return pitch

def index2key(index=49):
    """Given the SPN index, return the corresponding SPN key

    Example:

        >>> index2key(1)
        'A0'
        >>> index2key(4)
        'C1'
        >>> index2key(49)
        'A4'
        >>>

    """

    logger = logging.getLogger(__name__)

    assert isinstance(index, int)

    notes   = ("C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B")

    # shift the index around so that
    #    C0 --> -8
    #    A0 -->  1
    #    C1 -->  4
    notes_idx   = (index-3)%12 -1
    octave      = (index+8)//12

    # use the lookup table 'notes' to get the note name
    note        = notes[notes_idx]

    logger.debug("index %3i notes_idx %3i: %s%i" %(index, notes_idx, note, octave))

    spn_key = "%s%i" %(note, octave)

    return spn_key

__all__ = [
           'key2index',
           'index2key',
           ]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG')

    print('-- Done --')



"""
Created on 15 Feb 2015

@author: Ronny Andersson (ronny@andersson.tk)
@copyright: (c) 2015 Ronny Andersson
@license: MIT
"""

# standard library
from abc import ABCMeta, abstractmethod
import logging
import warnings

# external libraries
import numpy as np
try:
    import pyaudio
except ImportError:
    warnings.warn("PyAudio not found. Will not be able to create sndcard instances", category=ImportWarning)

# local libraries
from zignal import Audio, Noise

def list_devices():
    """List all available sound cards."""
    return PA.list_devices()

#===================================================================================================
# Abstract Base Class, inherit and implement the methods marked as @abstractmethod
#===================================================================================================
class _Device(object, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(__name__)

    def __str__(self):
        s  = '=======================================\n'
        s += 'classname        : %s\n' %self.__class__.__name__
        return s

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self):
        self._logger.debug("--- open")

    def close(self):
        self._logger.debug("--- close")

    @abstractmethod
    def play(self, x, **kwargs):
        """Play audio"""
        self._logger.debug("--- play")
        assert isinstance(x, Audio)

    @abstractmethod
    def rec(self, duration=None, channels=1, fs=96000, **kwargs):
        """Record audio"""
        self._logger.debug("--- rec")
        assert duration is not None, "Specify a duration (in seconds) to record for"

    @abstractmethod
    def play_rec(self, x, **kwargs):
        """Play and record audio"""
        self._logger.debug("--- play_rec")
        assert isinstance(x, Audio)

#===================================================================================================
# Stub class
#===================================================================================================
class Stub(_Device):
    """Stub device that can be dropped in anywhere as a fake sound card.
    The record methods will return noise. This is intended to be used
    during development when a real device would block.
    """

    def play(self, x, **kwargs):
        _Device.play(self, x, **kwargs)
        self._logger.warn("*** Stub play")

    def rec(self, duration=None, channels=1, fs=96000, **kwargs):
        _Device.rec(self, duration=duration, channels=channels, fs=fs, **kwargs)
        self._logger.warn("*** Stub rec")

        # fake a signal with white noise
        n = Noise(channels=channels, fs=fs, duration=duration, gaindb=-60)
        n.convert_to_float(targetbits=32)
        y = Audio(fs=fs, initialdata=n.samples)
        return y

    def play_rec(self, x, **kwargs):
        _Device.play_rec(self, x, **kwargs)
        self._logger.warn("*** Stub play_rec")

        # fake a signal with white noise
        n = Noise(channels=x.ch, fs=x.fs, nofsamples=x.nofsamples, gaindb=-60)
        n.convert_to_float(targetbits=32)
        y = Audio(fs=x.fs, initialdata=n.samples)
        return y

#===================================================================================================
# PyAudio (Portaudio) implementation
#===================================================================================================
class PA(_Device):
    """PyAudio wrapper. Uses the Audio base class as input signal and returns
    Audio instances after a recording. This implementation is using the blocking
    strategy."""

    #--------------------------------------------------------------------------------
    #    Portaudio frame
    #
    #    http://music.columbia.edu/pipermail/portaudio/2007-January/006615.html
    #
    #    A 'frame' is all data required at a snapshot in time, where snapshots
    #    are the same as samplerate. This means that 2 ch of 16bit audio is
    #    a frame of size 32 bits
    #
    #            +------------+------------+
    #    stereo  | 16bit ch1  | 16bit ch2  |
    #            +------------+------------+
    #
    #                         |
    #                         v
    #
    #            +-------------------------+
    #            |         4 bytes         |
    #            +-------------------------+
    #
    #    frame size calculation:
    #    2 bytes per sample, 2 channels --> 2*2=4 bytes
    #
    #    Another example:
    #    32 bit float, 8 channels --> 4 bytes per channel, 8 channels == 32 bytes
    #
    #            +-----+-----+-----+-----+-----+-----+-----+-----+
    #            | ch1 | ch2 | ch3 | ch4 | ch5 | ch6 | ch7 | ch8 |
    #            +-----+-----+-----+-----+-----+-----+-----+-----+
    #
    #            +-----------------------------------------------+
    #            |                   (32 bytes)                  |
    #            +-----------------------------------------------+
    #
    #    In other words, because we have audio data in a numpy vector
    #    where (rows,colums) --> (samples, channels) this means that each *row*
    #    is a frame.
    #
    #--------------------------------------------------------------------------------

    def __init__(self, device_out='default', device_in='default'):
        """Set the device_out and/or device_in string based on the name of the
        sound card. An int can also be used if the id is known beforehand. Note that
        the id can change when another sound card is detected, for example when a USB
        card is connected. The available sound cards can be found by calling list_devices()
        """
        _Device.__init__(self)

        self._device_out = device_out
        self._device_in  = device_in

        if isinstance(device_out, int):
            self._index_out = device_out
        else:
            self._index_out = self._get_id(name=device_out, find_output=True)

        if isinstance(device_in, int):
            self._index_in = device_in
        else:
            self._index_in = self._get_id(name=device_in,  find_output=False)

    def __str__(self):
        s  = _Device.__str__(self)
        s += 'portaudio        : %s %s\n'       %(pyaudio.get_portaudio_version(),
                                                  pyaudio.get_portaudio_version_text())
        s += 'pyaudio          : %s\n'          %pyaudio.__version__
        s += 'output device    : id %i, %s\n'   %(self._index_out, self._device_out)
        s += 'input device     : id %i, %s\n'   %(self._index_in,  self._device_in)
        return s

    def _get_id(self, name=None, find_output=True):
        """Find the id of the sound card that matches the string name"""
        retval = -1
        pa_get_id = pyaudio.PyAudio()

        try:
            if name=='default':
                if find_output:
                    device = pa_get_id.get_default_output_device_info()
                    if device['maxOutputChannels'] > 0:
                        self._device_out = device['name']
                        retval = device['index']

                else:
                    device = pa_get_id.get_default_input_device_info()
                    if device['maxInputChannels'] > 0:
                        self._device_in = device['name']
                        retval = device['index']
            else:
                for idx in range(pa_get_id.get_device_count()):
                    device = pa_get_id.get_device_info_by_index(idx)

                    if find_output:
                        if device['maxOutputChannels'] > 0:
                            if device['name'] == name:
                                retval = idx
                                break
                    else:
                        if device['maxInputChannels'] > 0:
                            if device['name'] == name:
                                retval = idx
                                break
        finally:
            pa_get_id.terminate()

        if retval == -1:
            s = "Device '%s' not found. Check the inputs and outputs arguments" %name
            print(s)
            try:
                print("Available devices: \n%s" %self.list_devices())
            finally:
                raise LookupError(s)

        return retval

    def open(self):
        """Open a PyAudio instance. This needs to be called before play(),
        play_rec() or rec() is called. This can be done in two ways:

            snd = PA()
            snd.open()
            try:
                snd.play(x)
            finally:
                snd.close()

        or use the 'with' statement:

            with PA() as snd:
                snd.play(x)

        """
        self._logger.debug("creating pyaudio instance")
        self.pa = pyaudio.PyAudio()

    def close(self):
        """Terminate the PyAudio instance. Must be called if open() has been called"""
        self.pa.terminate()
        self._logger.debug("pyaudio instance terminated")

    @classmethod
    def list_devices(cls):
        """Get a pretty string with all available sound cards.

        When using a portaudio instance, the id of the sound device needs
        to be known. This method is listing the available devices so that
        the id can be found.
        """

        s  = ''
        s += '--------------------------------------------------------------------\n'
        s += 'id out  in  def.fs   API            name\n'
        s += '--------------------------------------------------------------------\n'
        #--->| 0   2   2  44100.0  ALSA           Intel 82801AA-ICH: - (hw:0,0)
        pa_list_dev = pyaudio.PyAudio()
        try:
            for idx in range(pa_list_dev.get_device_count()):
                device = pa_list_dev.get_device_info_by_index(idx)
                s+='%2i %3i %3i %8.1f  %s %s\n' %(
                    device['index'],
                    device['maxOutputChannels'],
                    device['maxInputChannels'],
                    device['defaultSampleRate'],
                    pa_list_dev.get_host_api_info_by_index(device['hostApi'])['name'].ljust(len('Windows WASAPI')),
                    device['name'],
                    )
            s += '\n'
            s += 'default output device id: %i\n' %pa_list_dev.get_default_output_device_info()['index']
            s += 'default input  device id: %i\n' %pa_list_dev.get_default_input_device_info()['index']
            s += '--------------------------------------------------------------------\n'
        finally:
            pa_list_dev.terminate()

        return s

    def _data_format(self, x):
        """The data types in numpy needs to be mapped to the equivalent type in
        portaudio. This is an issue for 24 bit audio files since there isn't a
        24 bit data type in numpy. This is currently not implemented. There are
        some options on how to do this. We could for example use a 32 bit int and
        store the 24 bits either so that bits 1 to 8 is set to zeroes or so that
        bits 25 to 32 is set to zeros.
        """
        retval = None

        if x.samples.dtype == np.dtype(np.float32):
            self._logger.debug("pyaudio.paFloat32")
            retval = pyaudio.paFloat32
        elif x.samples.dtype == np.dtype(np.int16):
            self._logger.debug("pyaudio.paInt16")
            retval = pyaudio.paInt16
        elif x.samples.dtype == np.dtype(np.int32):
            self._logger.debug("pyaudio.paInt32")
            retval = pyaudio.paInt32
        else:
            raise NotImplementedError("Data type not understood: %s" %x.samples.dtype)

        return retval

    def _check_pow2(self, n):
        """Check that buffer size is a power of 2 (32, 64, ..., 1024, 2048, ...)"""
        check = 2**int(np.round(np.log2(n))) == n
        return check

    def _validate(self, frames_per_buffer):
        assert hasattr(self, "pa"), \
            "Call open() or use the 'with' statement before using play(), rec() or play_rec()"

        assert self._check_pow2(frames_per_buffer), \
            "Use a buffer size that is a power of 2 (1024, 2048, 4096, ...)"

        return True

    def _get_missing_frames(self, frames_per_buffer, length):
        """Calculate the number of frames missing to fill a buffer"""
        missing_frames = frames_per_buffer - (length%frames_per_buffer)

        self._logger.debug("frames per buffer : %i" %frames_per_buffer)
        self._logger.debug("missing frames    : %i" %missing_frames)

        return missing_frames


    def play(self, x, frames_per_buffer=1024):
        """Play audio. If dropouts or buffer underruns occur try different
        values for the frames_per_buffer variable."""

        _Device.play(self, x)
        self._validate(frames_per_buffer)

        missing_frames = self._get_missing_frames(frames_per_buffer, len(x))

        # generate silence to fill up missing frames
        pad = Audio(channels=x.ch, fs=x.fs, nofsamples=missing_frames, dtype=x.samples.dtype)

        # append the missing frames to a copy of the audio to be played. We now have
        # audio that can be split into complete (full) buffers
        cpy = Audio(fs=x.fs, initialdata=x.samples)
        cpy.concat(pad)

        assert len(cpy)%frames_per_buffer == 0

        stream = self.pa.open(format                = self._data_format(x),
                              channels              = x.ch,
                              rate                  = x.fs,
                              frames_per_buffer     = frames_per_buffer,
                              output_device_index   = self._index_out,
                              input                 = False,
                              output                = True,
                              )
        try:
            self._logger.info("play: start")
            counter = 0

            # split the audio into chunks the size of one buffer, so we can
            # iterate over the audio in chunksizes of the same size as one buffer
            it = iter(np.split(cpy.samples, len(cpy)/frames_per_buffer))
            try:
                while True:
                    chunk = next(it)
                    stream.write(chunk.tostring(), num_frames=frames_per_buffer)
                    counter += 1

            except StopIteration:
                pass

            finally:
                stream.stop_stream()

            self._logger.debug("chunks played  : %i"    %counter)
            self._logger.debug("samples played : %i"    %(counter*frames_per_buffer))
            self._logger.debug("duration       : %.3f"  %(counter*frames_per_buffer/x.fs))

        finally:
            self._logger.debug("play: close stream")
            stream.close()

        self._logger.info("play: done")

    def play_rec(self, x, frames_per_buffer=1024):
        """Play audio and record from input. If dropouts or buffer underruns occur
        try different values for the frames_per_buffer variable."""

        _Device.play_rec(self, x)
        self._validate(frames_per_buffer)

        missing_frames = self._get_missing_frames(frames_per_buffer, len(x))

        # generate silence to fill up missing frames
        pad = Audio(channels=x.ch, fs=x.fs, nofsamples=missing_frames, dtype=x.samples.dtype)

        # append the missing frames to a copy of the audio to be played. We now have
        # audio that can be split into complete (full) buffers
        cpy = Audio(fs=x.fs, initialdata=x.samples)
        cpy.concat(pad)

        assert len(cpy)%frames_per_buffer == 0

        rec = Audio(channels=cpy.ch, fs=cpy.fs, nofsamples=len(cpy), dtype=cpy.samples.dtype)

        stream = self.pa.open(format                = self._data_format(x),
                              channels              = x.ch,
                              rate                  = x.fs,
                              frames_per_buffer     = frames_per_buffer,
                              input_device_index    = self._index_in,
                              output_device_index   = self._index_out,
                              input                 = True,
                              output                = True,
                              )
        try:
            self._logger.info("play_rec: start")
            counter = 0

            # split the audio into chunks the size of one buffer, so we can
            # iterate over the audio in chunksizes of the same size as one buffer
            it_out = iter(np.split(cpy.samples, len(cpy)/frames_per_buffer))
            it_in  = iter(np.split(rec.samples, len(rec)/frames_per_buffer))
            try:
                while True:
                    chunk_out   = next(it_out)
                    chunk_in    = next(it_in)

                    stream.write(chunk_out.tostring(), num_frames=frames_per_buffer)

                    raw_1d      = np.fromstring(stream.read(frames_per_buffer),
                                                dtype=rec.samples.dtype)
                    # because we use an iterator chunk_in is a sliding window in the rec variable
                    chunk_in[:] = raw_1d.reshape((frames_per_buffer, rec.ch))

                    counter += 1

            except StopIteration:
                pass

            finally:
                stream.stop_stream()

            self._logger.debug("chunks played  : %i"    %counter)
            self._logger.debug("samples played : %i"    %(counter*frames_per_buffer))
            self._logger.debug("duration       : %.3f"  %(counter*frames_per_buffer/x.fs))

        finally:
            self._logger.debug("play_rec: close stream")
            stream.close()

        # remove the padding (empty frames) added to fill the last buffer. Trim
        # at the start, since we can treat that as latency.
        rec.trim(start=missing_frames, end=None)

        self._logger.debug("play_rec: trimmed %i samples from the start" %missing_frames)
        self._check_if_clipped(rec)
        self._logger.info("play_rec: done")

        return rec

    def rec(self, duration=None, channels=1, fs=96000, frames_per_buffer=1024, dtype=np.float32):
        """Record. If dropouts or buffer underruns occur try different
        values for the frames_per_buffer variable."""

        _Device.rec(self, duration=duration, channels=channels, fs=fs)
        self._validate(frames_per_buffer)

        missing_frames = self._get_missing_frames(frames_per_buffer, int(duration*fs))

        nofsamples = missing_frames+int(duration*fs)

        rec = Audio(channels=channels, fs=fs, nofsamples=nofsamples, dtype=dtype)

        assert len(rec)%frames_per_buffer == 0

        stream = self.pa.open(format                = self._data_format(rec),
                              channels              = rec.ch,
                              rate                  = rec.fs,
                              frames_per_buffer     = frames_per_buffer,
                              input_device_index    = self._index_in,
                              input                 = True,
                              output                = False,
                              )
        try:
            self._logger.info("rec: start")
            counter = 0

            # split the audio into chunks the size of one buffer, so we can
            # iterate over the audio in chunksizes of the same size as one buffer
            it_in = iter(np.split(rec.samples, len(rec)/frames_per_buffer))
            try:
                while True:
                    chunk_in    = next(it_in)
                    raw_1d      = np.fromstring(stream.read(frames_per_buffer),
                                                dtype=rec.samples.dtype)
                    # because we use an iterator chunk_in is a sliding window in the rec variable
                    chunk_in[:] = raw_1d.reshape((frames_per_buffer, rec.ch))

                    counter += 1

            except StopIteration:
                pass

            finally:
                stream.stop_stream()

            self._logger.debug("chunks recorded : %i" %counter)
            self._logger.debug("samples recorded: %i" %(counter*frames_per_buffer))
            self._logger.debug("duration        : %.3f" %(counter*frames_per_buffer/rec.fs))

        finally:
            self._logger.debug("rec: close stream")
            stream.close()

        # remove the padding (empty frames) added to fill the last buffer. Trim
        # at the start, since we can treat that as latency.
        rec.trim(start=missing_frames, end=None)

        self._logger.debug("rec: trimmed %i samples from the start" %missing_frames)
        self._check_if_clipped(rec)
        self._logger.info("rec: done")

        return rec

    def _check_if_clipped(self, rec):
        """check if the recording clipped, log the first clip for each channel"""

        clipped = False

        if np.issubdtype(rec.samples.dtype, np.floating):
            max_possible_positive_value = 1.0
        else:
            # integers used.
            # get the size of the integer type used, in bytes (2 for 16bit, 4 for 32bit)
            dt = np.dtype(rec.samples.dtype)

            # calculate the maximum possible postitive value. The maximum negative
            # value is max_possible_positive_value+1 (two's complement)
            max_possible_positive_value = 2**((8*dt.itemsize)-1) - 1

        self._logger.debug("maximum possible positive value: %i" %max_possible_positive_value)

        for i, peaks in enumerate(zip(rec.peak()[0], rec.peak()[1])):
            peak_val, peak_pos = peaks
            # abs(-32768) overflows in signed 16 bit, use long(...) in py2 to get a bigger data type
            if abs(int(peak_val)) >= max_possible_positive_value:
                clipped = True
                clip_position = peak_pos/rec.fs
                self._logger.warn("channel %02i clipped at %.3f" %(i+1, clip_position))

        return clipped

__all__ = [
           'list_devices',
           'PA',
           'Stub',
           ]

if __name__ == '__main__':
    logging.basicConfig(format='%(levelname)-7s: %(module)s.%(funcName)-15s %(message)s',
                        level='DEBUG',
                        )
    print(list_devices())
    print('++ End of script ++')
